#define _POSIX_C_SOURCE  200809L

#include "drivers/sx1239.h"

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

#include "utils/logger.h"

struct device {
	int tty;
	uint8_t radio;
	uint32_t bitrate;
	uint32_t bandwidth;
	int running;
	uint32_t part;
	uint32_t total;
};

static int tty_init(const char *tty)
{
	struct termios termios;
	int fd;

	if ((fd = open(tty, O_RDWR | O_NOCTTY)) < 0) {
		eerror("Failed to open");
		return -1;

	} else if (tcflush(fd, TCIOFLUSH) < 0) {
		eerror("Failed to flush");
		close(fd);
		return -1;

	} else if (tcgetattr(fd, &termios) < 0) {
		eerror("Failed to get terminal attributes");
		close(fd);
		return -1;

	} else if (cfsetispeed(&termios, B115200) < 0) {
		eerror("Failed to set input speed");
		close(fd);
		return -1;

	} else if (cfsetospeed(&termios, B115200) < 0) {
		eerror("Failed to set output speed");
		close(fd);
		return -1;
	}

	termios.c_cflag |= (CLOCAL | CREAD);
	termios.c_cflag &= ~PARENB;
	termios.c_cflag &= ~CSTOPB;
	termios.c_cflag &= ~CSIZE;
	termios.c_cflag |= CS8;
	termios.c_iflag &= ~(IXON | IXOFF | IXANY);
	termios.c_iflag &= ~(ICRNL | INLCR | IGNCR);
	termios.c_iflag |= (IGNBRK | IGNPAR);
	termios.c_oflag = 0;
	termios.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
	termios.c_cc[VMIN] = 0;
	termios.c_cc[VTIME] = 50;

	if (tcsetattr(fd, TCSANOW, &termios) < 0) {
		eerror("Failed to set terminal attributes");
		close(fd);
		return -1;

	} else if (tcflush(fd, TCIFLUSH) < 0) {
		eerror("Failed to flush");
		close(fd);
		return -1;
	}

	return fd;
}

void *sx1239_open(const char *tty)
{
	struct device *dev;

	if ((dev = malloc(sizeof(*dev))) == NULL) {
		eerror("Failed to allocate sx1239 device");
		return NULL;

	} else if ((dev->tty = tty_init(tty)) < 0) {
		error("Failed to initialize sx1239 tty");
		free(dev);
		return NULL;

	} else if (dprintf(dev->tty, "\r") < 0) {
		error("Failed to send first empty command");
		close(dev->tty);
		free(dev);
		return NULL;

	} else {
		dev->running = 0;
	}

	return dev;
}

int sx1239_close(void *device)
{
	struct device *dev = device;

	if (dev == NULL) {
		return 0;

	} else {
		sx1239_abort(device);
	}

	return close(dev->tty);
}

int sx1239_abort(void *device)
{
	struct device *dev = device;

	if (dev == NULL) {
		return 0;

	} else if (dev->running == 1 && dprintf(dev->tty, "%c", 3) < 0) {
		error("Failed to send SIGINT");
		return -1;

	} else {
		dev->running = 0;
	}

	return 0;
}

int sx1239_configure(void *device, uint8_t radio,
                     uint32_t bandwidth, uint32_t bitrate)
{
	struct device *dev = device;

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else {
		dev->radio = radio;
		dev->bitrate = bitrate;
		dev->bandwidth = bandwidth;
	}

	return 0;
}

static ssize_t nointr_read(int fd, void *buf, size_t length)
{
	size_t len;
	ssize_t l;

	for (len = 0; len < length;) {
		if ((l = read(fd, buf + len, length - len)) < 0) {
			if (errno == EINTR) {
				continue;

			} else {
				return -1;
			}

		} else {
			len += l;
		}
	}

	return len;
}

int sx1239_gecko_version(void *device, char *buffer, size_t length)
{
	struct device *dev = device;
	size_t len;
	int match;
	char c;

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else if (tcflush(dev->tty, TCIOFLUSH) < 0) {
		eerror("Failed to flush");
		return -1;

	} else if (dprintf(dev->tty, "printenv sw_version\r") < 0) {
		error("Failed to send command");
		return -1;
	}

	for (match = 0; match < 3;) {
		if (nointr_read(dev->tty, &c, 1) < 0) {
			eerror("Failed to read");
			return -1;

		} else if (match == 0 && c == ' ') {
			match = 1;

		} else if (match == 1 && c == '=') {
			match = 2;

		} else if (match == 2 && c == ' ') {
			match = 3;

		} else {
			match = 0;
		}
	}

	for (len = 0; len < length - 1; len++) {
		if (nointr_read(dev->tty, &c, 1) < 0) {
			eerror("Failed to read");
			return -1;

		} else if (c == '\r' || c == '\n') {
			break;

		} else {
			buffer[len] = c;
		}
	}

	buffer[len] = '\0';
	return 0;
}

static int create_histogram(struct device *dev, uint16_t *histogram,
                            uint32_t fstart, uint32_t fstep,
                            uint8_t steps, uint32_t samples,
                            void (*progress)(uint32_t part, uint32_t total))
{
	uint32_t fstop = fstart + fstep * (steps - 1);
	uint32_t step, sample, colon;
	char *b, buf[3] = { 0 };
	uint8_t rssi;

	if (tcflush(dev->tty, TCIOFLUSH) < 0) {
		eerror("Failed to flush");
		return -1;

	} else if (dprintf(dev->tty, "sx1239 fl %u %u %u %u %u %u %u\r",
	                   dev->radio + 1, fstart, fstop, fstep,
	                   dev->bandwidth, dev->bitrate, samples) < 0) {
		error("Failed to send command");
		return -1;
	}

	for (sample = 0; sample < samples && dev->running == 1; sample++) {
		for (colon = 0; colon < 2;) {
			if (nointr_read(dev->tty, buf, 1) < 0) {
				eerror("Failed to read");
				return -1;

			} else if (*buf == ':') {
				colon++;
			}
		}

		for (step = 0; step < steps; step++) {
			if (nointr_read(dev->tty, buf, 2) < 0) {
				eerror("Failed to read");
				return -1;
			}

			rssi = strtol(buf, &b, 16);

			if (b - buf != 2) {
				eerror("Failed to read rssi");
				return -1;

			} else if (rssi > SX1239_HISTO_VALUE_MAX) {
				warning("RSSI value out of range: %.1f dBm",
				        (float) - rssi);

			} else {
				histogram[step * SX1239_HISTO_STEP_NB + rssi]++;
				dev->part++;
			}

			if (progress != NULL) {
				progress(dev->part, dev->total);
			}
		}
	}

	return 0;
}

#define HISTO_MAX_STEP_NB               UINT8_MAX

uint16_t *sx1239_histogram(void *device, uint32_t fstart, uint32_t fstop,
                           uint32_t fstep, uint32_t samples,
                           void (*progress)(uint32_t part, uint32_t total))
{
	struct device *dev = device;
	uint32_t s, step_nb, steps = (fstop - fstart) / fstep + 1;
	uint16_t *h, *histogram;
	uint32_t f;

	if (dev == NULL) {
		error("Invalid device");
		return NULL;

	} else if ((histogram = calloc(steps * SX1239_HISTO_STEP_NB,
	                               sizeof(uint16_t))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;

	} else {
		dev->total = steps * samples;
		dev->part = 0;
		dev->running = 1;
	}

	for (s = 0, f = fstart, h = histogram; s < steps && dev->running == 1;) {
		step_nb = (steps - s > HISTO_MAX_STEP_NB) ?
		          HISTO_MAX_STEP_NB : steps - s;

		if (create_histogram(dev, h, f, fstep, step_nb, samples, progress) < 0) {
			free(histogram);
			return NULL;

		} else {
			h += step_nb * SX1239_HISTO_STEP_NB;
			f += step_nb * fstep;
			s += step_nb;
		}
	}

	if (progress != NULL) {
		progress(0, 0);
	}

	dev->running = 0;

	return histogram;
}
