#include <getopt.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "drivers/sx1239.h"
#include "utils/logger.h"

#include "spi_linuxdev.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"

#define FMIN                    862.0
#define FSTART                  863.0
#define FSTOP                   870.0
#define FMAX                    1020.0

#define FSTEP_MIN               0.0005
#define FSTEP                   0.2

#define FILENAME                "scan.csv"

#define SX1239_TTY              "/dev/ttyACM0"
#define SX1239_RADIO            0
#define SX1239_BITRATE          600

static bool running = true;
static bool graceful_exit = true;
static void *sx1239 = NULL;

static bool lna_started = false;
static int spi = -1;

static void signal_handler(int signal)
{
	switch (signal) {
	case SIGQUIT:
		graceful_exit = false;
		/* fallthrough */
	case SIGINT:
		/* fallthrough */
	case SIGTERM:
		running = false;
		sx1239_abort(sx1239);
		break;
	}
}

static int handle_signals(void)
{
	struct sigaction signal_action;

	signal_action.sa_handler = signal_handler;
	sigemptyset(&signal_action.sa_mask);
	signal_action.sa_flags = 0;

	if (sigaction(SIGINT, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGINT");
		return -1;

	} else if (sigaction(SIGQUIT, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGQUIT");
		return -1;

	} else if (sigaction(SIGTERM, &signal_action, NULL) < 0) {
		eerror("Failed to handle SIGTERM");
		return -1;
	}

	return 0;
}

static void print_histogram_freq(uint16_t *histogram, uint32_t freq,
                                 uint16_t samples, uint8_t basic)
{
	uint16_t h, s;
	float r, a;
	int32_t i;

	for (i = SX1239_HISTO_VALUE_MAX, a = 0.0, s = 0; i >= 0; i--) {
		r = -i;
		h = histogram[i];

		if (h != 0) {
			a = (a * s + r * h) / (s + h);
			s += h;
		}
	}

	if (basic == 0) {
		fprintf(stdout, "%10u Hz %-6.1f ", freq, a);

	} else {
		fprintf(stdout, "%u %.1f", freq, a);
	}

	for (i = SX1239_HISTO_VALUE_MAX; i >= 0 && basic == 0; i -= 2) {
		h = histogram[i - 0] + histogram[i - 1];

		if (h == 0) {
			putc(' ', stdout);

		} else if (h < samples / 100) {
			putc('.', stdout);

		} else {
			putc('|', stdout);
		}
	}

	putc('\n', stdout);
}

static void print_histogram(uint16_t *histogram,
                            uint32_t fstart, uint32_t fstop, uint32_t fstep,
                            uint16_t samples, uint8_t basic)
{
	uint16_t *h;
	uint32_t f;

	if (basic == 0) {
		fprintf(stdout, "\n%-20s | %6.1f dBm %52.1f dBm |\n",
		        "Histogram scale",
		        (float) - SX1239_HISTO_VALUE_MAX, 0.0);
	}

	for (h = histogram, f = fstart;
	     f <= fstop;
	     f += fstep, h += SX1239_HISTO_STEP_NB) {
		print_histogram_freq(h, f, samples, basic);
	}
}

static int save_histogram_freq(FILE *file, uint16_t *histogram,
                               uint32_t freq, uint8_t strip_zero)
{
	int32_t i;

	if (fprintf(file, "%u", freq) < 0) {
		error("Failed to print histogram header");
		return -1;
	}

	for (i = SX1239_HISTO_VALUE_MAX; i >= 0; i--) {
		if (strip_zero == 0 || histogram[i] != 0) {
			if (fprintf(file, ",%.1f,%u",
			            (float) - i, histogram[i]) < 0) {
				error("Failed to print histogram value");
				return -1;
			}
		}
	}

	if (fprintf(file, "\n") < 0) {
		error("Failed to print histogram footer");
		return -1;
	}

	return 0;
}

static int save_histogram(FILE *file, uint16_t *histogram,
                          uint32_t fstart, uint32_t fstop, uint32_t fstep,
                          uint8_t strip_zero)
{
	uint16_t *h;
	uint32_t f;

	for (h = histogram, f = fstart;
	     f <= fstop;
	     f += fstep, h += SX1239_HISTO_STEP_NB) {
		if (save_histogram_freq(file, h, f, strip_zero) < 0) {
			return -1;
		}
	}

	return 0;
}

static void print_progress(uint32_t part, uint32_t total)
{
	if (total == 0) {
		fprintf(stdout, "\r%7s\r", " ");
		fflush(stdout);

	} else {
		fprintf(stdout, "\r%5.1f %%", 100.0 * part / total);
		fflush(stdout);
	}
}

static int spi_read(uint8_t header, uint16_t address, uint8_t *data,
                    uint32_t size, uint8_t *status)
{
	return spi_linuxdev_read(header, spi, address, data, size, status);
}

static int spi_write(uint8_t header, uint16_t address, const uint8_t *data,
                     uint32_t size, uint8_t *status)
{
	return spi_linuxdev_write(header, spi, address, data, size, status);
}

static int lna_start(char *spidev)
{
	sx1301ar_board_cfg_t cfg = sx1301ar_init_board_cfg();
	int32_t global_en, clk32m_en;

	if (spi_linuxdev_open(spidev, 1, &spi) < 0) {
		eerror("Failed to open spidev: %s", spidev);
		return -1;
	}

	cfg.spi_read = &spi_read;
	cfg.spi_write = &spi_write;

	/* Dummy parameters, just need to set SPI callbacks */
	cfg.board_type = BRD_MASTER;
	cfg.nb_chip = 2;
	cfg.rx_freq_hz = 868E6;
	cfg.rx_bw_hz = 4E6;
	cfg.rf_chain[0].rx_enable = true;

	if (sx1301ar_conf_board(0, &cfg) < 0) {
		error("Failed to configure board: %s",
		      sx1301ar_err_message(sx1301ar_errno));
		return -1;

	} else if (sx1301ar_reg_chip_r(0, 0, CREG_GLOBAL_EN, &global_en) < 0) {
		error("Failed to read register: %s",
		      sx1301ar_err_message(sx1301ar_errno));
		return -1;

	} else if (sx1301ar_reg_chip_r(0, 0, CREG_CLK32M_EN, &clk32m_en) < 0) {
		error("Failed to read register: %s",
		      sx1301ar_err_message(sx1301ar_errno));
		return -1;

	} else if (global_en == 1 && clk32m_en == 1) {
		return 0;

	} else {
		lna_started = true;
	}

	if (sx1301ar_reg_reset(0) < 0) {
		warning("Failed to reset board: %s",
		        sx1301ar_err_message(sx1301ar_errno));
		return -1;

	} else if (sx1301ar_reg_brd_w(0, BREG_FE_RX_A_EN, 1) < 0) {
		warning("Failed to configure radio: %s",
		        sx1301ar_err_message(sx1301ar_errno));
		return -1;

	} else if (sx1301ar_reg_brd_w(0, BREG_FE_RX_B_EN, 1) < 0) {
		warning("Failed to configure radio: %s",
		        sx1301ar_err_message(sx1301ar_errno));
		return -1;
	}

	return 0;
}

static int lna_stop(void)
{
	int ret = 0;

	if (lna_started == true && sx1301ar_reg_reset(0) < 0) {
		ret = -1;
	}

	if (spi_linuxdev_close(spi) < 0) {
		ret = -1;
	}

	return ret;
}

int main(int argc, char **argv)
{
	struct timespec start, end;
	int status = EXIT_FAILURE;
	int opt, iopt, itm;

	char *tty = NULL;
	char *spidev = NULL;

	char *filename = NULL;
	FILE *file = NULL;

	uint32_t fstart = FSTART * 1E6;
	uint32_t fstop = FSTOP * 1E6;
	uint32_t fstep = FSTEP * 1E6;
	double fvector[3];

	uint16_t samples = 1024;
	uint16_t *histogram = NULL;

	uint8_t radio = SX1239_RADIO;
	uint32_t bandwidth = 0;
	uint32_t bitrate = SX1239_BITRATE;

	uint8_t start_lna = 0;
	uint8_t strip_zero = 0;
	uint8_t basic = 0;
	uint8_t force = 0;

	void (*progress)(uint32_t, uint32_t) = &print_progress;
	char gversion[16];

	const struct option opts[] = {
		{"device",     required_argument,  NULL, 'd'},
		{"frequency",  required_argument,  NULL, 'f'},
		{"samples",    required_argument,  NULL, 's'},
		{"output",     required_argument,  NULL, 'o'},
		{"radio",      required_argument,  NULL, 'r'},
		{"bandwidth",  required_argument,  NULL, 'b'},
		{"bitrate",    required_argument,  NULL, 'B'},
		{"start-lna",  required_argument,  NULL, 'l'},
		{"strip-zero", no_argument,        NULL, '0'},
		{"basic",      no_argument,        NULL,  0 },
		{"force",      no_argument,        NULL,  0 },
		{"version",    no_argument,        NULL, 'v'},
		{"help",       no_argument,        NULL, 'h'},
		{NULL,         0,                  NULL,  0 },
	};

	/* Get start time */
	if (clock_gettime(CLOCK_MONOTONIC, &start) < 0) {
		error("Failed to get start time");
		goto exit;
	}

	/* Manage options */
	while ((opt = getopt_long(argc, argv,
	                          "d:f:s:o:r:b:B:l:0vh", opts, &iopt)) != EOF) {
		switch (opt) {
		case 'd':
			free(tty);

			if ((tty = strdup(optarg)) == NULL) {
				eerror("Failed to duplicate tty");
				goto exit;
			}

			break;

		case 'f':
			if ((itm = sscanf(optarg, "%lf:%lf:%lf", &fvector[0],
			                  &fvector[1], &fvector[2])) < 0) {
				eerror("Failed to parse frequency vector");
				goto exit;

			}

			switch (itm) {
			case 1:
				fvector[2] = fvector[0];
				fvector[1] = FSTEP;
				break;

			case 2:
				fvector[2] = fvector[1];
				fvector[1] = FSTEP;
				break;
			}

			if (fvector[0] > fvector[2]) {
				error("FSTART is greater than FSTOP");
				goto exit;

			} else if (fvector[0] < FMIN && force == 0) {
				error("FSTART is lower than %f", FMIN);
				goto exit;

			} else if (fvector[2] > FMAX && force == 0) {
				error("FSTOP is greater than %f", FMAX);
				goto exit;

			} else if (fvector[1] < FSTEP_MIN && force == 0) {
				error("FSTEP is lower than %f", FSTEP_MIN);
				goto exit;
			}

			fstart = fvector[0] * 1E6;
			fstep = fvector[1] * 1E6;
			fstop = fvector[2] * 1E6;
			break;

		case 's':
			if (sscanf(optarg, "%hu", &samples) < 1) {
				eerror("Failed to read samples");
				goto exit;

			} else if (samples == 0) {
				error("Samples number cannot be null");
				goto exit;
			}

			break;

		case 'o':
			free(filename);

			if ((filename = strdup(optarg)) == NULL) {
				eerror("Failed to duplicate filename");
				goto exit;
			}

			break;

		case 'r':
			if (sscanf(optarg, "%hhu", &radio) < 1) {
				eerror("Failed to read radio");
				goto exit;

			} else if (radio > 1) {
				error("Radio is out of range");
				goto exit;
			}

			break;

		case 'b':
			if (sscanf(optarg, "%u", &bandwidth) < 1) {
				eerror("Failed to read bandwidth");
				goto exit;
			}

			break;

		case 'B':
			if (sscanf(optarg, "%u", &bitrate) < 1) {
				eerror("Failed to read bitrate");
				goto exit;
			}

			break;

		case 'l':
			start_lna = 1;

			free(spidev);

			if ((spidev = strdup(optarg)) == NULL) {
				eerror("Failed to duplicate spidev");
				goto exit;
			}

			break;

		case '0':
			strip_zero = 1;
			break;

		case 0:
			if (strcmp("basic", opts[iopt].name) == 0) {
				progress = NULL;
				basic = 1;

			} else if (strcmp("force", opts[iopt].name) == 0) {
				force = 1;
			}

			break;

		case 'v':
			fprintf(stdout, "klk_spectral_scan " PACKAGE_VERSION "\n");
			status = EXIT_SUCCESS;
			goto exit;

		default:
			fprintf(stderr, "Usage: klk_spectral_scan [OPTIONS]...\n"
			        "\n"
			        "-d, --device DEVICE               "
			        "    Serial port to use.\n"
			        "-f, --frequency FSTART:FSTEP:FSTOP"
			        "    Scan frequency vector (in MHz).\n"
			        "-s, --samples NUMBER              "
			        "    Number of RSSI samples per scan.\n"
			        "-o, --output FILENAME             "
			        "    Output filename.\n"
			        "-r, --radio RADIO                 "
			        "    Use the specified radio (0 or 1).\n"
			        "-b, --bandwidth BANDWIDTH         "
			        "    Use the specified bandwidth (in Hz).\n"
			        "-B, --bitrate BITRATE             "
			        "    Use the specified bit rate (in bps).\n"
			        "-l, --start-lna SPIDEV            "
			        "    Start LNA using SPIDEV if not already done.\n"
			        "-0, --strip-zero                  "
			        "    Don't save zero values.\n"
			        "    --basic                       "
			        "    Print script friendly outputs.\n"
			        "-v, --version                     "
			        "    Display version.\n"
			        "-h, --help                        "
			        "    Display this help screen.\n");
			goto exit;
		}
	}

	if (bandwidth == 0) {
		bandwidth = fstep;
	}

	if (start_lna == 1 && lna_start(spidev) < 0) {
		error("Failed to start LNA");
		goto exit;
	}

	/* Handle signals */
	if (handle_signals() < 0) {
		error("Failed to handle signals");
		goto exit;
	}

	/* Open output file */
	if (filename == NULL && (filename = strdup(FILENAME)) == NULL) {
		eerror("Failed to duplicate filename");
		goto exit;

	} else if ((file = fopen(filename, "w")) == NULL) {
		eerror("Failed to open output file");
		goto exit;
	}

	/* Initialize sniffer */
	if (tty == NULL && (tty = strdup(SX1239_TTY)) == NULL) {
		eerror("Failed to duplicate tty");
		goto exit;

	} else if ((sx1239 = sx1239_open(tty)) == NULL) {
		error("Failed to open sx1239");
		goto exit;

	} else if (sx1239_configure(sx1239, radio, bandwidth, bitrate) < 0) {
		error("Failed to configure sx1239");
		goto exit;

	} else if (sx1239_gecko_version(sx1239, gversion, sizeof(gversion)) < 0) {
		error("Failed to get sx1239 gecko version");
		goto exit;
	}

	info("Start frequency: %17u Hz", fstart);
	info("Stop frequency:  %17u Hz", fstop);
	info("Step frequency:  %17u Hz", fstep);
	info("RSSI samples:    %20u", samples);
	info("Bandwidth:       %17u Hz", bandwidth);
	info("Bit rate:        %16u bps", bitrate);
	info("Radio:           %20u", radio);
	info("Serial port:     %20s", tty);

	if (start_lna == 1) {
		info("SPI device port: %20s", spidev);
	}

	info("Gecko version:   %20s", gversion);
	info("Output filename: %20s", filename);

	if ((histogram = sx1239_histogram(sx1239, fstart, fstop,
	                                  fstep, samples, progress)) == NULL) {
		error("Failed to get histogram");
		goto exit;

	} else if (save_histogram(file, histogram, fstart, fstop,
	                          fstep, strip_zero) < 0) {
		error("Failed to save histogram");
		goto exit;

	} else {
		print_histogram(histogram, fstart, fstop, fstep, samples, basic);
	}

	/* Display footer message */
	if (basic == 0) {
		if (clock_gettime(CLOCK_MONOTONIC, &end) < 0) {
			info("Scan %s", running ? "finished" : "exited");

		} else {
			if (end.tv_nsec < start.tv_nsec) {
				end.tv_sec -= start.tv_sec + 1;
				end.tv_nsec -= start.tv_nsec - 1E9;

			} else {
				end.tv_sec -= start.tv_sec;
				end.tv_nsec -= start.tv_nsec;
			}

			info("Scan %s in %.3f second%s",
			     running ? "finished" : "exited",
			     end.tv_sec + end.tv_nsec / 1E9,
			     end.tv_sec ? "s" : "");
		}
	}

	status = graceful_exit ? EXIT_SUCCESS : EXIT_FAILURE;

exit:   /* Clean up */

	if (sx1239 != NULL) {
		sx1239_close(sx1239);
	}

	if (file != NULL) {
		fclose(file);
	}

	if (start_lna == 1) {
		lna_stop();
	}

	free(histogram);
	free(filename);
	free(spidev);
	free(tty);
	exit(status);
}
