/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Auxiliary functions for LoRa Gateway v2

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/


#ifndef _SX1301_ARRAY_AUX_H
#define _SX1301_ARRAY_AUX_H

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <stdint.h>     /* C99 types */
#include <sys/time.h>   /* gettimeofday, structtimeval */

/* -------------------------------------------------------------------------- */
/* --- PUBLIC CONSTANTS AND MACROS ------------------------------------------ */

/* Debug log filters */
#define DEBUG_INFO      0
#define DEBUG_MSG_FIFO  0   /* Debug FPGA/DSP FIFO binary data */
#define DEBUG_MERGE     0   /* Debug SX1301 packet merge */
#define DEBUG_MATCH     0   /* Debug fine timestamp matching */
#define DEBUG_MATCH_ERR 0   /* Debug fine timestamp matching errors */
#define DEBUG_TIME      0   /* Debug timing issues, time conversions etc... */
#define DEBUG_AES       0   /* Debug fine timestamp decryption */
#define DEBUG_GPS       0   /* Debug GPS command */
#define DEBUG_DSP       0   /* Debug DSP functions */
#define DEBUG_HISTORY   0   /* Debug RX history */
#define DEBUG_TEMP_COMP 0   /* Debug temperature compensation */
#define DEBUG_PERF      0   /* Debug timing performances: level [0..4] */
#define DEBUG_LBT       0   /* Debug Listen-Before-Talk */
#define LOG_FTIME_STAT  0   /* Print fine timestamp statistics */

/* Debug log frequencies */
#define LOG_FTIME_FREQ  50  /* How often FTIME_STAT are printed */

#define MSG(args...) printf(args) /* message that is destined to the user */
#define MSG_DEBUG(FLAG, fmt, ...)                                                                         \
            do  {                                                                                         \
                if (FLAG)                                                                                 \
                    fprintf(stdout, fmt, ##__VA_ARGS__); \
            } while (0)

/* Get a particular bit value from a byte */
/* b: any byte
   p: index >=0
   n: number of bits >=1 */
/**
@brief Get a particular bit value from a byte
@param b [in]   Any byte from which we want a bit value
@param p [in]   Position of the bit in the byte [0..7]
@param n [in]   Number of bits we want to get
@return The value corresponding the requested bits
*/
#define TAKE_N_BITS_FROM( b, p, n ) ((b) >> (p)) & ((1 << (n)) - 1)

/**
@brief Substract struct timeval values
@param a [in]   struct timeval a
@param b [in]   struct timeval b
@param b [out]  struct timeval resulting from (a - b)
*/
#define TIMER_SUB( a, b, result )                                              \
    do  {                                                                      \
        (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                          \
        (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                       \
        if ((result)->tv_usec < 0) {                                           \
            --(result)->tv_sec;                                                \
            (result)->tv_usec += 1000000;                                      \
        }                                                                      \
    } while (0)

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS PROTOTYPES ------------------------------------------ */
/** @warning This module has no internal state. */

/**
@brief Wait for a certain time
@param delay_us  Duration of the wait in micro-seconds
*/
void sx1301ar_wait_us( uint32_t delay_us );

/**
@brief Wait for a certain time
@param delay_ms  Duration of the wait in milli-seconds
*/
void sx1301ar_wait_ms( uint32_t delay_ms );

/**
@brief return current system time in number of milli-seconds since epoch
*/
unsigned long long sx1301ar_timestamp( void );

/**
@brief Calculate 8-bit fletcher checksum of given data
@param data     Pointer to byte array containing data on which calculate the checksum
@param size     Size of the byte array
@param crc_a    First byte of the checksum calculated
@param crc_b    Second byte of the checksum calculated
*/
void sx1301ar_calculate_fletcher_checksum( uint8_t * data, size_t size, uint8_t * crc_a, uint8_t * crc_b );

/**
@brief Convert a big endian 4-bytes byte array into little endian unsigned int
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding unsigned int
*/
uint32_t bytes_be_to_uint32_le( uint8_t * bytes );

/**
@brief Convert a little endian 4-bytes byte array into little endian unsigned int
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding unsigned int
*/
uint32_t bytes_le_to_uint32_le( uint8_t * bytes );

/**
@brief Convert a little endian 4-bytes byte array into little endian unsigned short
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding unsigned short
*/
uint16_t bytes_le_to_uint16_le( uint8_t * bytes );

/**
@brief Convert a big endian 4-bytes byte array into little endian unsigned short
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding unsigned short
*/
uint16_t bytes_be_to_uint16_le( uint8_t * bytes );

/**
@brief Convert a big endian 4-bytes byte array into little endian signed short
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding signed short
*/
int16_t bytes_be_to_int16_le( uint8_t * bytes );

/**
@brief Convert a big endian 8-bytes byte array into little endian unsigned long integer
@param bytes  Pointer to the 8 bytes byte-array to be converted
@return The corresponding unsigned long integer
*/
uint64_t bytes_be_to_uint64_le( uint8_t * bytes );

/**
@brief Convert a 4-bytes byte array into a 32-bits float
@param bytes  Pointer to the 4 bytes byte-array to be converted
@return The corresponding float
*/
float bytes_to_float( uint8_t * bytes );

/**
@brief Transform IF in Hz for a SX1301 into FPGA register value
@param if_freq  IF frequency, in Hz (signed)
@param p_freq_bias frequency error due to round number
@return register value (13 bit useful)
*/
int16_t chip_if_freq2reg( int32_t if_freq, int16_t *p_freq_bias );

/**
@brief Transform FPGA register value into IF in Hz for a SX1301
@param if_reg  register value (13 bit useful)
@return IF frequency, in Hz (signed)
*/
int32_t chip_if_reg2freq( int16_t if_reg );

/**
@brief Transform IF in Hz for a channel into SX1301 register value
@param if_freq  IF frequency, in Hz (signed)
@param p_freq_bias frequency error due to round number
@return register value (13 bit useful)
*/
int16_t chan_if_freq2reg( int32_t if_freq, int16_t *p_freq_bias );

/**
@brief Record the current time, for measure start
@param tm Pointer to the current time value
*/
void _meas_time_start( struct timeval *tm );

/**
@brief Measure the ellapsed time since given time
@param debug_level  debug print debug level to be used
@param start_time   start time of the measure to be used
@param str          string to be used for debug print
*/
void _meas_time_stop( int debug_level, struct timeval start_time, const char *str );

#endif

/* --- EOF ------------------------------------------------------------------ */
