/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    Utility to get Altera FPGA chip ID

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <unistd.h>     /* getopt */
#include <string.h>
#include <errno.h>

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"

#include "spi_linuxdev.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define LINUXDEV_PATH_DEFAULT   "/dev/spidev32766.0"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

static void usage( void );

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
    int i, ret;
    uint8_t tmp_buff[4];
    uint32_t fpga_id_lsb;
    uint32_t fpga_id_msb;
    sx1301ar_btype_t board_type = BRD_MASTER;
    char arg_s[64];

    /* SPI interfaces */
    const char spi1_path_default[] = LINUXDEV_PATH_DEFAULT;
    const char * spi1_path = spi1_path_default;

    /* Configuration structures */
    sx1301ar_board_cfg_t cfg_brd = sx1301ar_init_board_cfg( );

    /* Parse command line options */
    while( (i = getopt( argc, argv, "hd:b:" )) != -1 )
    {
        switch( i )
        {
        case 'h':
            usage( );
            return EXIT_SUCCESS;
            break;

        case 'd': /* -d <mandatory path>  use Linux SPI device driver */
            if( optarg != NULL )
            {
                spi1_path = optarg;
            }
            break;

        case 'b': /* -b <str> board type */
            ret = sscanf( optarg, "%10s", arg_s );
            if( (ret != 1) || ((strcmp(arg_s, "MASTER") != 0) && (strcmp(arg_s, "SLAVE"))) )
            {
                printf( "ERROR: Invalid board type (MASTER, SLAVE)\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                if( strcmp(arg_s, "MASTER") == 0)
                {
                    board_type = BRD_MASTER;
                }
                else
                {
                    board_type = BRD_SLAVE;
                }
            }
            break;

        default:
            printf( "ERROR: argument parsing options\n" );
            usage( );
            return EXIT_FAILURE;
        }
    }

    /* Open SPI link */
    ret = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( ret != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i (%s : %s)\n", ret, strerror(errno), spi1_path );
        return EXIT_FAILURE;
    }

    cfg_brd.spi_read = &spi1_read;
    cfg_brd.spi_write = &spi1_write;
    cfg_brd.board_type = board_type;
    /* Dummy parameters, just need to set SPI callbacks */
    cfg_brd.nb_chip = 1;
    cfg_brd.rx_freq_hz = 868E6;
    cfg_brd.rx_bw_hz = 7E6;
    cfg_brd.full_duplex = false;
    ret = sx1301ar_conf_board( 0, &cfg_brd );
    if( ret != 0 )
    {
        printf( "ERROR: sx1301ar_conf_board failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    /* Initialize board (check VERSION registers) */
    ret = sx1301ar_reg_init( 0, cfg_brd.nb_chip, NULL );
    if( ret != 0 )
    {
        printf( "ERROR: Failed to communicate to board. %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    if( board_type == BRD_SLAVE )
    {
        printf( "INFO: board SLAVE\n" );
    }
    else
    {
        printf( "INFO: board MASTER\n" );
    }

    /* Reset FPGA */
    sx1301ar_reg_brd_w( 0, BREG_SOFT_RESET, 1 );
    sx1301ar_reg_brd_w( 0, BREG_SOFT_RESET, 0 );
    sx1301ar_wait_us( 1000 );

    /* Get Chip ID */
    ret = sx1301ar_reg_brd_rb( 0, BREG_ALT_CHIP_ID_LSB, tmp_buff, 4 );
    if ( ret != 0 ) {
        printf ("ERROR: Failed to read FPGA Chip ID LSB\n");
    } else {
        fpga_id_lsb = tmp_buff[0];
        fpga_id_lsb |= tmp_buff[1] << 8;
        fpga_id_lsb |= tmp_buff[2] << 16;
        fpga_id_lsb |= tmp_buff[3] << 24;
        printf ("CHIP ID (LSB): 0x%08x\n", fpga_id_lsb);
    }
    ret = sx1301ar_reg_brd_rb( 0, BREG_ALT_CHIP_ID_MSB, tmp_buff, 4 );
    if ( ret != 0 ) {
        printf ("ERROR: Failed to read FPGA Chip ID MSB\n");
    } else {
        fpga_id_msb = tmp_buff[0];
        fpga_id_msb |= tmp_buff[1] << 8;
        fpga_id_msb |= tmp_buff[2] << 16;
        fpga_id_msb |= tmp_buff[3] << 24;
        printf ("CHIP ID (MSB): 0x%08x\n", fpga_id_msb);
    }

    sx1301ar_reg_reset( 0 );

    /* Close SPI link */
    ret = spi_linuxdev_close( spi1_context );
    if( ret != 0 )
    {
        printf( "ERROR: closing SPI failed, returned %i\n", ret );
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void usage( void )
{
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " -h  print this help\n" );
    printf( " -d <path>  use Linux SPI device driver\n" );
    printf( "            => default path: " LINUXDEV_PATH_DEFAULT "\n" );
    printf( " -b <str>   board type [MASTER, SLAVE]\n" );
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */
