/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Listen Before Talk basic test application

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDANCIES --------------------------------------------------------- */

/* fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdbool.h>    /* bool type */
#include <stdio.h>      /* printf fprintf sprintf fopen fputs */

#include <signal.h>     /* sigaction */
#include <unistd.h>     /* getopt access */
#include <stdlib.h>     /* rand */

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_reg.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_radio.h"

#include "spi_linuxdev.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#define LINUXDEV_PATH_DEFAULT       "/dev/spidev32766.0"

#define DEFAULT_RADIO_RX_FREQ       866500000 /* Hz */
#define LBT_RSSI_SHIFT              142  /* to convert RSSI target in dBm to register value */
#define LBT_RSSI_TARGET_MIN         -105 /* dBm */
#define LBT_RSSI_TARGET_MAX         -42 /* dBm */
#define LBT_NUMBER_OF_CHANNELS_MAX  16
#define LBT_CHANNEL_FREQ_STEP       200E3 /* Hz */
#define LBT_SCAN_TIME1              128 /* µs */
#define LBT_SCAN_TIME2              5000 /* µs */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES (GLOBAL) ------------------------------------------- */

/* signal handling variables */
struct sigaction sigact; /* SIGQUIT&SIGINT&SIGTERM signal handling */
static int exit_sig = 0; /* 1 -> application terminates cleanly (shut down hardware, close open files, etc) */
static int quit_sig = 0; /* 1 -> application terminates without shutting down the hardware */

int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DECLARATION ---------------------------------------- */

static void sig_handler( int sigio );

void usage( void );

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char **argv )
{
    int i, x;
    int xi = 0;

    /* SPI interfaces */
    const char spi1_path_default[] = LINUXDEV_PATH_DEFAULT;
    const char * spi1_path = spi1_path_default;

    /* in/out variables */
    uint32_t rx_freq_hz = DEFAULT_RADIO_RX_FREQ;
    double f1 = 0.0;
    uint32_t f_start = 0; /* in Hz */
    int8_t rssi_target = -80; /* in dBm */
    uint8_t nb_channels = 8;
    uint16_t scan_time = LBT_SCAN_TIME2;

    /* program variables */
    uint32_t timestamp;
    int32_t val, val2;
    int16_t lbt_if_reg;

    /* parse command line options */
    while( (i = getopt( argc, argv, "hd:f:r:n:s:c:" )) != -1 )
    {
        switch( i )
        {
            case 'h':
                usage( );
                return EXIT_FAILURE;

            case 'd': /* -d <mandatory path>  use Linux SPI device driver */
                if( optarg != NULL )
                {
                    spi1_path = optarg;
                }
                break;

            case 'f':
                i = sscanf( optarg, "%lf", &f1 );
                if( (i != 1) || (f1 < 30.0) || (f1 > 3000.0) )
                {
                    MSG( "ERROR: Invalid LBT start frequency\n" );
                    usage( );
                    return EXIT_FAILURE;
                }
                else
                {
                    f_start = (uint32_t)((f1*1e6) + 0.5);/* .5 Hz offset to get rounding instead of truncating */
                }
                break;

            case 'c':
                i = sscanf( optarg, "%lf", &f1 );
                if( (i != 1) || (f1 < 30.0) || (f1 > 3000.0) )
                {
                    MSG( "ERROR: Invalid radio RX center frequency\n" );
                    usage( );
                    return EXIT_FAILURE;
                }
                else
                {
                    rx_freq_hz = (uint32_t)((f1*1e6) + 0.5);/* .5 Hz offset to get rounding instead of truncating */
                }
                break;

            case 'r':
                i = sscanf( optarg, "%i", &xi );
                if( (i != 1) || (xi < LBT_RSSI_TARGET_MIN) || (xi > LBT_RSSI_TARGET_MAX) )
                {
                    MSG( "ERROR: rssi_target must be b/w %i & %i\n", LBT_RSSI_TARGET_MIN, LBT_RSSI_TARGET_MAX );
                    usage( );
                    return EXIT_FAILURE;
                }
                else
                {
                    rssi_target = xi;
                }
                break;

            case 'n':
                i = sscanf( optarg, "%i", &xi );
                if( (i != 1) || (xi < 1) || (xi > LBT_NUMBER_OF_CHANNELS_MAX) )
                {
                    MSG( "ERROR: number of LBT channels must be b/w 1 & %d \n", LBT_NUMBER_OF_CHANNELS_MAX );
                    usage( );
                    return EXIT_FAILURE;
                }
                else
                {
                    nb_channels = xi;
                }
                break;

            case 's':
                i = sscanf( optarg, "%i", &xi );
                if( (i != 1) || ((xi != LBT_SCAN_TIME1) && (xi != LBT_SCAN_TIME2)) )
                {
                    MSG( "ERROR: scan time must be %d or %d µs\n", LBT_SCAN_TIME1, LBT_SCAN_TIME2 );
                    usage( );
                    return EXIT_FAILURE;
                }
                else
                {
                    scan_time = xi;
                }
                break;

            default:
                MSG( "ERROR: argument parsing use -h option for help\n" );
                usage( );
                return EXIT_FAILURE;
        }
    }

    /* Sanity check */
    if( f_start == 0 )
    {
        MSG( "ERROR: LBT start frequency must be set\n" );
        usage( );
        return EXIT_FAILURE;
    }

    MSG( "INFO: Starting LoRa Gateway v2 LBT test\n" );

    /* configure signal handling */
    sigemptyset( &sigact.sa_mask );
    sigact.sa_flags = 0;
    sigact.sa_handler = sig_handler;
    sigaction( SIGQUIT, &sigact, NULL );
    sigaction( SIGINT, &sigact, NULL );
    sigaction( SIGTERM, &sigact, NULL );

    /* Open SPI link */
    x = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( x != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i\n", x );
        return EXIT_FAILURE;
    }

    /* Configure SPI callbacks for REG module */
    x = sx1301ar_reg_setup( 0, BRD_MASTER, &spi1_read, &spi1_write );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_reg_setup failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    /* Get board information */
    sx1301ar_reg_brd_r( 0, BREG_VERSION, &val);
    MSG( "INFO: FPGA VERSION = %d\n", val );

    /* Configure LBT */
    sx1301ar_reg_brd_w( 0, BREG_LBT_FILTER_1_GAIN, 15 );
    sx1301ar_reg_brd_w( 0, BREG_LBT_FILTER_1_DC_NOTCH_EN, 0 );
    sx1301ar_reg_brd_w( 0, BREG_LBT_FILTER_2_GAIN, 13 );
    sx1301ar_reg_brd_w( 0, BREG_LBT_SELECT_RF_PATH, 0 );
    sx1301ar_reg_brd_w( 0, BREG_LBT_RSSI_START_DELAY, 500 );

    /* Configure LBT channels */
    sx1301ar_reg_brd_w( 0, BREG_LBT_NB_CHANNEL, nb_channels-1 );
    for( i = 0; i < nb_channels; i++ )
    {
        lbt_if_reg = chip_if_freq2reg( (int32_t)rx_freq_hz - (int32_t)(f_start + (i * LBT_CHANNEL_FREQ_STEP)), NULL );
        sx1301ar_reg_brd_w( 0, BREG_LBT_CHANNEL_0_IF_FREQ + i, lbt_if_reg );
    }

    /* Enable all channels, and rely on NB_CHANNEL */
    sx1301ar_reg_brd_w( 0, BREG_LBT_CHANNEL_ENABLE, 0xFFFF );

    /* Set LBT RSSI target threshold */
    sx1301ar_reg_brd_w( 0, BREG_LBT_RSSI_TARGET, (int32_t)(rssi_target + LBT_RSSI_SHIFT) );

    /* Set LBT scan time for all channels */
    sx1301ar_reg_brd_w( 0, BREG_LBT_RSSI_READ_DELAY, 10 ); /* scan time = RSSI_READ_DELAY * READ_RSSI_TIMEx */
    sx1301ar_reg_brd_w( 0, BREG_LBT_NB_READ_RSSI_TIME1, 13 ); /* 128µs */
    sx1301ar_reg_brd_w( 0, BREG_LBT_NB_READ_RSSI_TIME2, 500 ); /* 5000µs */
    if( scan_time == LBT_SCAN_TIME1 )
    {
        sx1301ar_reg_brd_w( 0, BREG_LBT_CHANNEL_TIME_MODE, 0x0 );
    }
    else
    {
        sx1301ar_reg_brd_w( 0, BREG_LBT_CHANNEL_TIME_MODE, 0xFFFF );
    }

    /* Start LBT state machine */
    sx1301ar_reg_brd_w( 0, BREG_START_LBT, 1 );

    /* Read back LBT config */
    sx1301ar_reg_brd_r( 0, BREG_LBT_NB_CHANNEL, &val );
    if( val != (nb_channels-1) )
    {
        MSG( "ERROR: Wrong number of LBT channels configured: %i (should be %u)\n", val, nb_channels-1 );
        return EXIT_FAILURE;
    }
    MSG( "INFO: Number of LBT channels: %u\n", val+1);
    sx1301ar_reg_brd_r( 0, BREG_LBT_RSSI_TARGET, &val );
    if( val != (rssi_target + LBT_RSSI_SHIFT) )
    {
        MSG( "ERROR: Wrong RSSI target configured: %i (should be %i)\n", val, rssi_target + LBT_RSSI_SHIFT );
        return EXIT_FAILURE;
    }
    MSG( "INFO: RSSI target: %idBm\n", val-LBT_RSSI_SHIFT);
    sx1301ar_reg_brd_r( 0, BREG_LBT_CHANNEL_TIME_MODE, &val );
    if( val != ((scan_time == LBT_SCAN_TIME1) ? 0x0 : 0xFFFF) )
    {
        MSG( "ERROR: Wrong scan time configured: %i (should be %u)\n", val, (scan_time == LBT_SCAN_TIME1) ? 0x0 : 0xFFFF );
        return EXIT_FAILURE;
    }
    MSG( "INFO: Scan time: 0x%X\n", val);
    for( i = 0; i < nb_channels; i++ )
    {
        sx1301ar_reg_brd_r( 0, BREG_LBT_CHANNEL_0_IF_FREQ + i, &val );
        val2 = chip_if_reg2freq( (int16_t)val );
        MSG( "INFO: LBT channel %02d frequency: %.1f MHz\n", i, (double)((uint32_t)val2 + rx_freq_hz) / 1E6 );
    }

    /* Checking SX1301 and FPGA counters sync */
    sx1301ar_reg_chip_r( 0, SX1301AR_GPS_TIMER_CHIP, CREG_TIMESTAMP, &val );
    sx1301ar_reg_brd_r( 0, BREG_SX1301_TIMESTAMP_CPY_GPS, &val2 );
    if( abs(val - val2) > 1 ) /* tolerate 1µs */
    {
        MSG( "ERROR: SX1301 and FPGA counters are not synced (SX1301:%u - FPGA:%u)\n", (uint32_t)val, (uint32_t)val2 );
        return EXIT_FAILURE;
    }

    /* Start test */
    while( (quit_sig != 1) && (exit_sig != 1) )
    {
        MSG( " ~~~~~~~~~~~~\n" );
        /* Get current SX1301 timestamp (1MHz) */
        sx1301ar_reg_brd_r( 0, BREG_SX1301_TIMESTAMP_CPY, &val );
        timestamp = (uint32_t)val;
        MSG( " TIMESTAMP = %u\n", timestamp );

        for( i = 0; i < nb_channels; i++ )
        {
            /* Select LBT i */
            sx1301ar_reg_brd_w( 0, BREG_LBT_SELECT_CH, i );

            /* Get last instant when the selected i was free */
            sx1301ar_reg_brd_r( 0, BREG_LBT_TIMESTAMP, &val );
            timestamp = (uint32_t)val;
            MSG( " TIMESTAMP_CH%d = %u\n", i, timestamp );
        }

        sx1301ar_wait_ms( 400 );
    }

    /* Stop LBT state machine */
    sx1301ar_reg_brd_w( 0, BREG_START_LBT, 0 );

    /* Close SPI link */
    x = spi_linuxdev_close( spi1_context );
    if( x != 0 )
    {
        printf( "ERROR: closing SPI failed, returned %i\n", x );
        return EXIT_FAILURE;
    }
    printf( "PASS: SPI link closed\n" );

    MSG( "INFO: Exiting LoRa Gateway v2 LBT test successfully\n" );

    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void sig_handler( int sigio )
{
    if( sigio == SIGQUIT )
    {
        quit_sig = 1;;
    }
    else if( (sigio == SIGINT) || (sigio == SIGTERM) )
    {
        exit_sig = 1;
    }
}

/* describe command line options */
void usage( void )
{
    printf( "Available options:\n" );
    printf( " -h print this help\n" );
    printf( " -d <path>  use Linux SPI device driver\n" );
    printf( "            => default path: " LINUXDEV_PATH_DEFAULT "\n" );
    printf( " -f <float> start frequency in MHz\n" );
    printf( " -c <float> radio RX center frequency in MHz, default:%.1lf\n", DEFAULT_RADIO_RX_FREQ/1E6 );
    printf( " -n <uint>  number of LBT channels [0..%i]\n", LBT_NUMBER_OF_CHANNELS_MAX );
    printf( " -r <int>   target RSSI in dBm used to detect if the channel is clear or not [%i..%i]\n", LBT_RSSI_TARGET_MIN, LBT_RSSI_TARGET_MAX );
    printf( " -s <uint>  scan time in µs to be set for all LBT channels [%d,%d]\n", LBT_SCAN_TIME1, LBT_SCAN_TIME2 );
}

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */

