/*

Description:
    Utility to get Altera FPGA chip ID and firmware version

*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <unistd.h>     /* getopt */
#include <string.h>
#include <errno.h>
#include <time.h>

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"
#include "sx1301ar_dsp.h"

#include "spi_linuxdev.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define LINUXDEV_PATH_DEFAULT   "/dev/spidev0.0"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

static void usage( void );

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
  int i, ret;
  uint8_t tmp_buff[4];
  uint32_t cid_lsb = 0;
  uint32_t cid_msb = 0;
  int32_t swver = 0;
  int wait_on_reset = 0;

  /* SPI interfaces */
  const char spi1_path_default[] = LINUXDEV_PATH_DEFAULT;
  const char * spi1_path = spi1_path_default;

  /* Configuration structures */
  sx1301ar_board_cfg_t cfg_brd = sx1301ar_init_board_cfg();

  /* Parse command line options */
  while((i = getopt(argc, argv, "hd:w:")) != -1)
  {
    switch(i)
    {
      case 'h':
        usage();
        return EXIT_SUCCESS;
        break;

      case 'd': /* -d <optional path>  use Linux SPI device driver */
        if(optarg != NULL)
        {
          spi1_path = optarg;
        }
        break;

      case 'w': /* -w  wait and loop on FPGA reset indefinitely */
        sscanf( optarg, "%i", &wait_on_reset );
        break;

      default:
        printf("ERROR: argument parsing options\n");
        usage();
        return EXIT_FAILURE;
    }
  }

  /* Open SPI link @1MHz */
  ret = spi_linuxdev_open(spi1_path, 1, &spi1_context);
  if(ret != 0)
  {
    printf("ERROR: opening SPI failed, returned %i (%s : %s)\n", ret, strerror(errno), spi1_path);
    return EXIT_FAILURE;
  }

START_CONFIG:

  cfg_brd.spi_read = &spi1_read;
  cfg_brd.spi_write = &spi1_write;
  /* Dummy parameters, just need to set SPI callbacks */
  cfg_brd.board_type = BRD_MASTER;
  cfg_brd.nb_chip = 2;
  cfg_brd.rx_freq_hz = 868E6;
  cfg_brd.rx_bw_hz = 7E6;
  cfg_brd.rf_chain[0].rx_enable = true;
  ret = sx1301ar_conf_board(0, &cfg_brd);
  if(ret != 0)
  {
    printf("ERROR: sx1301ar_conf_board failed; %s\n", sx1301ar_err_message(sx1301ar_errno));
    return EXIT_FAILURE;
  }

  /* Reset FPGA */
  sx1301ar_reg_brd_w(0, BREG_SOFT_RESET, 1);
  sx1301ar_reg_brd_w(0, BREG_SOFT_RESET, 0);
  sx1301ar_wait_us(1000);

  /* Get Chip ID */
  ret = sx1301ar_reg_brd_rb(0, BREG_ALT_CHIP_ID_LSB, tmp_buff, 4);
  if(ret != 0)
  {
    printf("ERROR: Failed to read FPGA Chip ID LSB\n");
  }
  else
  {
    cid_lsb = tmp_buff[0];
    cid_lsb |= tmp_buff[1] << 8;
    cid_lsb |= tmp_buff[2] << 16;
    cid_lsb |= tmp_buff[3] << 24;
  }
  ret = sx1301ar_reg_brd_rb(0, BREG_ALT_CHIP_ID_MSB, tmp_buff, 4);
  if(ret != 0)
  {
    printf("ERROR: Failed to read FPGA Chip ID MSB\n");
  }
  else
  {
    cid_msb = tmp_buff[0];
    cid_msb |= tmp_buff[1] << 8;
    cid_msb |= tmp_buff[2] << 16;
    cid_msb |= tmp_buff[3] << 24;
  }

  ret = sx1301ar_reg_brd_r(0, BREG_VERSION, &swver);
  printf("fpga_cid = 0x%08X%08X\n", cid_msb, cid_lsb);
  printf("fpga_swver = %d\n", swver);

  if(wait_on_reset != 0)
  {
      sx1301ar_wait_us(wait_on_reset * 1000);
      printf("[%ld]\n",time(NULL));
      goto START_CONFIG;
  }

  for( i = 0; i < 2; i++ )
  {
    ret = sx1301ar_reg_chip_r(0, i, CREG_VERSION, &swver);
    if(ret < 0) swver = 0;
    ret = sx1301ar_reg_chip_r(0, i, CREG_CHIP_ID, (int32_t *) &cid_lsb);
    if(ret < 0) cid_lsb = 0;

    printf("Sx1301_%u_cid = %u\n", (i+1), cid_lsb);
    printf("Sx1301_%u_ver = %d\n", (i+1), swver);
  }

  for( i = 0; i < 2; i++ )
  {
    ret = gwloc_dsp_boot(0, i, 2);
#if 0
    if(ret == 0)
    {
      sx1301ar_reg_brd_r( 0, BREG_DSP_MIC_STATUS, &val );
      // mask for corresponding DSP (cf. sx1301ar_start )
      val = val >> (i*4);
      if( (val & 0x0F) == 5) ret = 0;
      else ret = -1;
    }
#endif
    printf("DSP_%u = %s\n", (i+1), (ret == 0)?"OK":"ERROR");
  }

  /* Stop the DSPs */
  gwloc_dsp_stop(0, 0, 1);
  gwloc_dsp_stop(0, 1, 1);

  sx1301ar_reg_reset(0);

  /* Close SPI link */
  ret = spi_linuxdev_close(spi1_context);
  if(ret != 0)
  {
    printf("ERROR: closing SPI failed, returned %i\n", ret);
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void usage( void )
{
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " -h  print this help\n" );
    printf( " -d <optional path>  use Linux SPI device driver\n" );
    printf( " -w <sleeptime_ms>   loop on FPGA reset all <sleeptime> ms \n" );
    printf( "  => default path: " LINUXDEV_PATH_DEFAULT "\n" );
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */
