/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    LoRa Gateway v2 packet logger

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdbool.h>    /* bool type */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <unistd.h>     /* getopt access */
#include <signal.h>     /* sigaction */
#include <string.h>     /* memset */
#include <time.h>       /* time clock_gettime strftime gmtime clock_nanosleep */
#include <pthread.h>    /* pthread */
#include <termios.h>    /* tcflush */
#include <fcntl.h>      /* open */
#include <errno.h>      /* error messages */

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_gps.h"

#include "spi_linuxdev.h"

#include "parson.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

#define JSON_CONF_DEFAULT       "config.json"
#define JSON_SX1301AR_CONF_OBJ  "SX1301_array_conf"
#define JSON_GATEWAY_CONF_OBJ   "gateway_conf"

#define LINUXDEV_PATH_DEFAULT   "/dev/spidev32766.0"
#define GPSTTY_PATH_DEFAULT     "/dev/ttyS4"

#define N_SAT_MIN       2 /* minimum acceptable satellite number */
#define HDOP_MAX        10.0 /* maximum acceptable HDOP */
#define FAILED_SYNC_MAX 3 /* number of failed GPS sync that force a resync */
#define REF_VALIDITY    60 /* number of seconds a time reference is considered valid */

/* -------------------------------------------------------------------------- */
/* --- CUSTOM TYPES --------------------------------------------------------- */

typedef struct
{
    uint64_t gw_id; /*!> LoRa gateway 64b ID (eg. MAC address) */
    char gw_id_str[17]; /*!> gateway ID, as a text string */
    int log_rotate_period; /*!> log file rotation period, in seconds (disabled if <= 0) */
}
gw_conf_t;

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* Signal handling variables */
static int exit_sig = 0; /* 1 -> application terminates cleanly (shut down hardware, close open files, etc) */
static int quit_sig = 0; /* 1 -> application terminates without shutting down the hardware */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* Hardware access control and correction */
static pthread_mutex_t mx_concent = PTHREAD_MUTEX_INITIALIZER; /* control access to the concentrator */
static pthread_mutex_t mx_tref = PTHREAD_MUTEX_INITIALIZER; /* control access to the time reference */

/* Multi-Boards */
static int board_nb = 1;

/* Gateway configuration */
static int rf_chain_nb = 0;

/* Timing */
static sx1301ar_tref_t time_ref;
static bool time_ref_valid = false;

/* Terminal attributes backup */
static struct termios ttyopt_restore;

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

/* Threads */
static void * thread_gps( const void * arg );

static void sig_handler( int sigio );

static void usage( void );

static int parse_sx1301ar_configuration( JSON_Object * brd_conf_obj, uint8_t brd );
static int parse_gateway_configuration( JSON_Object * gw_conf_obj, gw_conf_t * gwc );

static FILE * open_log( const gw_conf_t * gwc, time_t log_time );

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
    static struct sigaction sigact; /* SIGQUIT&SIGINT&SIGTERM signal handling */

    int i, j; /* loop and temporary variables */
    int x; /* return code for SX1301ar driver functions */
    uint8_t board; /* board iterator */

    /* SPI interfaces */
    const char spi1_path_default[] = LINUXDEV_PATH_DEFAULT;
    const char * spi1_path = spi1_path_default;

    /* GPS variables */
    bool enable_gps = false;
    const char gps_path_default[] = GPSTTY_PATH_DEFAULT;
    const char * gps_path = gps_path_default;
    int fd_gps = -1; /* file descriptor for the GPS TTY */
    struct termios ttyopt; /* serial port options */
    time_t ref_age;

    /* Threads ID */
    pthread_t thrid_gps;

    /* Gateway configuration variables */
    gw_conf_t gwc =
    {
        .gw_id = 0,
        .log_rotate_period = 3600 /* by default, rotation every hour */
    };

    /* Configuration file */
    const char defaut_conf_fname[] = JSON_CONF_DEFAULT;
    const char * conf_fname = defaut_conf_fname; /* pointer to a string we won't touch */
    JSON_Value * root_val;
    JSON_Object * root = NULL;
    JSON_Object * conf = NULL;
    JSON_Array * conf_array = NULL;

    /* Clock and log rotation management */
    time_t now_time;
    time_t log_start_time;
    FILE * log_file = NULL;
    int time_check = 0; /* variable used to limit the number of calls to time( ) function */
    unsigned long pkt_in_log = 0; /* count the number of packet written in each log file */
    unsigned long nb_pkt_crc_ok = 0;

    /* Allocate memory for packet fetching and processing */
    sx1301ar_rx_pkt_t rxbuf[SX1301AR_MAX_PKT_NB]; /* array to receive packet returned by fetch( ) */
    uint8_t nb_pkt; /* number of packets returned by fetch( ) */
    sx1301ar_rx_pkt_t * p; /* pointer to one RX packet structure */

    /* System time variables */
    struct timespec fetch_time;
    struct tm * fetch_tm;
    char fetch_timestamp[30];
    char fetch_timestamp_gps[30];

    /* Temperature variables */
    uint8_t u8_tempcode = 0;
    int16_t s16_temp_deg = 0;

    /* Parse command line options */
    while( (i = getopt( argc, argv, "hd:c:g:" )) != -1 )
    {
        switch( i )
        {
        case 'h':
            usage( );
            return EXIT_SUCCESS;
            break;

        case 'd': /* -d <path>  use Linux SPI device driver */
            if( optarg != NULL )
            {
                spi1_path = optarg;
            }
            break;

        case 'g': /* -g <optional path>  use GPS receiver for synchronization */
            enable_gps = true;
            if( optarg != NULL )
            {
                gps_path = optarg;
            }
            break;

        case 'c':
            conf_fname = optarg;
            break;

        default:
            printf( "ERROR: argument parsing options, use -h option for help\n" );
            usage( );
            return EXIT_FAILURE;
        }
    }

    /* Start message */
    printf( "+++ Start of packet logger for 'SX1301 array' concentrator +++\n" );
    printf( "SX1301 array library version string: %s\n", sx1301ar_version_info( 0, NULL, NULL ) );

    /* Open SPI link */
    x = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( x != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i\n", i );
        return EXIT_FAILURE;
    }

    /* Configure SPI devices to master/slave modes (Needed for ATMEL CPU only)
         SPI 0: HOST <-> FPGA
         SPI 1: HOST/DSP <-> Flash
    */
    x = spi_set_mode( 0, SPI_MODE_MASTER );
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", i );
        return EXIT_FAILURE;
    }
    x = spi_set_mode( 1, SPI_MODE_SLAVE ); /* So that DSP can access the flash for booting */
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", i );
        return EXIT_FAILURE;
    }

    if( enable_gps == true )
    {
        /* Init time reference (global val) */
        time_ref = sx1301ar_init_tref( );

        /* Open TTY device for GPS */
        fd_gps = open( gps_path, O_RDWR | O_NOCTTY );
        if( fd_gps <= 0 )
        {
            printf( "ERROR: opening TTY port for GPS failed - %s (%s)\n", strerror(errno), gps_path );
            return EXIT_FAILURE;
        }

        /* Save current TTY serial port parameters */
        x = tcgetattr( fd_gps, &ttyopt_restore );
        if( x != 0 )
        {
            printf( "ERROR: failed to get GPS serial port parameters\n" );
            return EXIT_FAILURE;
        }

        /* Get TTY serial port parameters */
        x = tcgetattr( fd_gps, &ttyopt );
        if( x != 0 )
        {
            printf( "ERROR: failed to get GPS serial port parameters\n" );
            return EXIT_FAILURE;
        }

        /* Update TTY terminal parameters */
        /* The following configuration should allow to:
            - Get ASCII NMEA messages
            - Get UBX binary messages
            - Send UBX binary commands
            Note: as binary data have to be read/written, we need to disable
                  various character processing to avoid loosing data */
        /* Control Modes */
        ttyopt.c_cflag |= CLOCAL; /* local connection, no modem control */
        ttyopt.c_cflag |= CREAD; /* enable receiving characters */
        ttyopt.c_cflag |= CS8; /* 8 bit frames */
        ttyopt.c_cflag &= ~PARENB; /* no parity */
        ttyopt.c_cflag &= ~CSTOPB; /* one stop bit */
        /* Input Modes */
        ttyopt.c_iflag |= IGNPAR;  /* ignore bytes with parity errors */
        ttyopt.c_iflag &= ~ICRNL;  /* do not map CR to NL on input*/
        ttyopt.c_iflag &= ~IGNCR;  /* do not ignore carriage return on input */
        ttyopt.c_iflag &= ~IXON;   /* disable Start/Stop output control */
        ttyopt.c_iflag &= ~IXOFF;  /* do not send Start/Stop characters */
        /* Output Modes */
        ttyopt.c_oflag = 0;        /* disable everything on output as we only write binary */
        /* Local Modes */
        ttyopt.c_lflag &= ~ICANON; /* disable canonical input - cannot use with binary input */
        ttyopt.c_lflag &= ~ISIG;   /* disable check for INTR, QUIT, SUSP special characters */
        ttyopt.c_lflag &= ~IEXTEN; /* disable any special control character */
        ttyopt.c_lflag &= ~ECHO;   /* do not echo back every character typed */
        ttyopt.c_lflag &= ~ECHOE;  /* does not erase the last character in current line */
        ttyopt.c_lflag &= ~ECHOK;  /* do not echo NL after KILL character */

        /* Settings for non-canonical mode:
           read will block for until the lesser of VMIN or requested chars have been received */
        ttyopt.c_cc[VMIN]  = 1; /* byte granularity at init time to ensure we get all UBX responses */
        ttyopt.c_cc[VTIME] = 0;

        /* Set new TTY serial ports parameters */
        x = tcsetattr( fd_gps, TCSANOW, &ttyopt );
        if( x != 0 )
        {
            printf( "ERROR: failed to update GPS serial port parameters\n" );
            return EXIT_FAILURE;
        }

        /* Display message and flush */
        printf( "GPS enabled for synchronization using device %s\n", gps_path );
        tcflush( fd_gps, TCIOFLUSH );
    }

    /* Parse configuration file */
    if( access( conf_fname, R_OK ) == 0 )
    {
        printf( "INFO: found configuration file %s\n", conf_fname );
        root_val = json_parse_file_with_comments( conf_fname );
        root = json_value_get_object( root_val );
        if( root == NULL )
        {
            printf( "ERROR: %s id not a valid JSON file\n", conf_fname );
            return EXIT_FAILURE;
        }

        conf_array = json_object_get_array( root, JSON_SX1301AR_CONF_OBJ );
        if( conf_array != NULL)
        {
            /* Point to JSON object containing sx1301ar concentrator configuration for one board */
            board_nb = (int)json_array_get_count( conf_array );
            for( i = 0; i < board_nb; i++ )
            {
                if( i >= SX1301AR_MAX_BOARD_NB)
                {
                    printf( "ERROR: board %d not supported, skip it\n", i );
                    break;
                }
                conf = json_array_get_object( conf_array, i );
                if( conf == NULL )
                {
                    printf( "ERROR: configuration file does not contain a JSON object named %s\n", JSON_SX1301AR_CONF_OBJ );
                    return EXIT_FAILURE;
                }

                printf( "INFO: parsing concentrator parameters for board #%d\n", i );
                x = parse_sx1301ar_configuration( conf, i );
                if( x != 0 )
                {
                    printf( "ERROR: failed to parse configuration %s from JSON file\n", JSON_SX1301AR_CONF_OBJ );
                    return EXIT_FAILURE;
                }
            }
        }

        /* Point to JSON object containing gateway configuration */
        conf = json_object_get_object( root, JSON_GATEWAY_CONF_OBJ );
        if( conf == NULL )
        {
            printf( "ERROR: configuration file does not contain a JSON object named %s\n", JSON_GATEWAY_CONF_OBJ );
            return EXIT_FAILURE;
        }
        printf( "INFO: parsing gateway parameters\n" );
        x = parse_gateway_configuration( conf, &gwc );
        if( x != 0 )
        {
            printf( "ERROR: failed to parse configuration %s from JSON file\n", JSON_GATEWAY_CONF_OBJ );
            return EXIT_FAILURE;
        }

        /* Free JSON parsing structure */
        json_value_free( root_val );
    }
    else
    {
        printf( "ERROR: failed to find any configuration file named %s\n", conf_fname );
        return EXIT_FAILURE;
    }

    /* Transform the MAC address into a string */
    sprintf( gwc.gw_id_str, "%08X%08X", (uint32_t)(gwc.gw_id >> 32), (uint32_t)(gwc.gw_id & 0xFFFFFFFF) );

    /* Start the concentrator */
    x = sx1301ar_start( board_nb );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_start failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( board_nb );
        return EXIT_FAILURE;
    }
    printf( "INFO: concentrator started\n" );

    /* If enabled, spawn the thread in charge of GPS synchronization */
    if( enable_gps == true )
    {
        x = pthread_create( &thrid_gps, NULL, (void * (*)(void *))thread_gps, (void *)&fd_gps );
        if( x != 0 )
        {
            printf( "ERROR: [main] impossible to create GPS thread\n" );
            return EXIT_FAILURE;
        }
    }

    /* Opening log file and writing CSV header*/
    time( &now_time );
    log_start_time = now_time;
    log_file = open_log( &gwc, log_start_time );
    if( log_file == NULL )
    {
        sx1301ar_stop( board_nb );
        return EXIT_FAILURE;
    }

    if( gwc.log_rotate_period <= 0)
    {
        printf( "INFO: log rotation disabled\n" );
    }
    else
    {
        printf( "INFO: log file will be rotated every %i seconds\n", gwc.log_rotate_period );
    }

    /* Configure signal handling */
    sigemptyset( &sigact.sa_mask );
    sigact.sa_flags = 0;
    sigact.sa_handler = sig_handler;
    sigaction( SIGQUIT, &sigact, NULL );
    sigaction( SIGINT, &sigact, NULL );
    sigaction( SIGTERM, &sigact, NULL );

    /* Main loop */
    board = -1;
    while( (quit_sig != 1) && (exit_sig != 1) )
    {
        /* Set next board or rotate */
        board++;
        if( board >= board_nb )
        {
            board = 0;
        }

        if( enable_gps == true )
        {
            pthread_mutex_lock( &mx_tref );
            ref_age = time( NULL ) - time_ref.systime;
            time_ref_valid = (ref_age >= 0) && (ref_age < REF_VALIDITY);
            pthread_mutex_unlock( &mx_tref );
        }

        /* Fetch packets */
        pthread_mutex_lock( &mx_concent );
        x = sx1301ar_fetch( board, rxbuf, ARRAY_SIZE( rxbuf ), &nb_pkt );
        pthread_mutex_unlock( &mx_concent );
        clock_gettime( CLOCK_REALTIME, &fetch_time );
        if( x != 0 )
        {
            printf( "ERROR: sx1301ar_fetch failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
            return EXIT_FAILURE;
        }

        if( nb_pkt == 0 )
        {
            sx1301ar_wait_us( 10 * 1000 );
        }
        else
        {
            /* Generate fetch timestamp string */
            fetch_tm = gmtime( &(fetch_time.tv_sec) );
            sprintf( fetch_timestamp, "%04i-%02i-%02i %02i:%02i:%02i.%03liZ", (fetch_tm->tm_year)+1900, (fetch_tm->tm_mon)+1, fetch_tm->tm_mday, fetch_tm->tm_hour, fetch_tm->tm_min, fetch_tm->tm_sec, (fetch_time.tv_nsec)/1000000 ); /* ISO 8601 format */

            /* Log received packets */
            for( i = 0; i < nb_pkt; ++i )
            {
                p = &rxbuf[i];

                if( time_ref_valid == true)
                {
                    /* convert packet timestamp to UTC absolute time */
                    pthread_mutex_lock( &mx_tref );
                    x = sx1301ar_cnt2utc(time_ref, p->count_us, &fetch_time);
                    pthread_mutex_unlock( &mx_tref );
                    if( x == 0 )
                    {
                        fetch_tm = gmtime( &(fetch_time.tv_sec) );
                        sprintf( fetch_timestamp_gps, "%04i-%02i-%02i %02i:%02i:%02i.%03liZ", (fetch_tm->tm_year)+1900, (fetch_tm->tm_mon)+1, fetch_tm->tm_mday, fetch_tm->tm_hour, fetch_tm->tm_min, fetch_tm->tm_sec, (fetch_time.tv_nsec)/1000000 ); /* ISO 8601 format */
                        /* "sys time (UTC)", empty "GPS timestamp" */
                        fprintf( log_file, "\"%s\",\"%s\",", fetch_timestamp, fetch_timestamp_gps );
                    }
                    else
                    {
                        /* "sys time (UTC)", empty "GPS timestamp" */
                        fprintf( log_file, "\"%s\",,", fetch_timestamp );
                    }
                }
                else
                {
                    /* "sys time (UTC)", empty "GPS timestamp" */
                    fprintf( log_file, "\"%s\",,", fetch_timestamp );
                }

                /* "board", "counter timestamp", "frequency" */
                fprintf( log_file, "%u,%010u,%u,", board, p->count_us, p->freq_hz );

                /* "channel nb", "RSSI" */
                for( j = 0; j < rf_chain_nb; j++ )
                {
                    if( p->rsig[j].is_valid == true )
                    {
                        fprintf( log_file, "%u,%.0f,", p->rsig[j].chan, p->rsig[j].rssi_chan );
                    }
                    else
                    {
                        /* leave the fields empty */
                        fprintf( log_file, ",," );
                    }
                }

                /* "status" */
                switch( p->status )
                {
                case STAT_CRC_OK:  fputs( "\"CRC OK\",", log_file ); nb_pkt_crc_ok++; break;
                case STAT_CRC_ERR: fputs( "\"CRC ERR\",", log_file ); break;
                case STAT_NO_CRC:  fputs( "\"NO CRC\",", log_file ); break;
                default: fputs( "\"?\",", log_file ); break;
                }

                /* "size" */
                fprintf( log_file, "%u,", p->size );

                if( p-> modulation == MOD_LORA )
                {
                    /* "modulation", "BW (kHz)", "SF", empty "data rate", "CR" */
                    fprintf( log_file, "\"LoRa\",%i,%i,,%i,", (int)(sx1301ar_bw_enum2nb( p->bandwidth )/1000), sx1301ar_sf_min_enum2nb( p->modrate ), sx1301ar_cr_enum2nb( p->coderate ) );
                    /* "SNR" */
                    for( j = 0; j < rf_chain_nb; j++ )
                    {
                        if( p->rsig[j].is_valid == true )
                        {
                            fprintf( log_file, "%+.1f,", p->rsig[j].snr );
                        }
                        else
                        {
                            /* leave the fields empty */
                            fprintf( log_file, "," );
                        }
                    }
                }
                else if( p->modulation == MOD_FSK )
                {
                    /* "modulation", empty "BW (kHz)" & "SF", "data rate", empty "CR" & "SNR"  */
                    fprintf( log_file, "\"FSK\",,,%6li,,,", (long)p->modrate );
                }
                else
                {
                    /* "modulation", empty "BW (kHz)" & "SF" & "data rate" & "CR" & "SNR" */
                    fputs( "\"ERROR\",,,,,,", log_file );
                }

                /* fine timestamps */
                for( j = 0; j < rf_chain_nb; j++ )
                {
                    if( (p->rsig[j].fine_received == true) && (time_ref_valid == true) )
                    {
                        fprintf( log_file, "%09d,", p->rsig[j].fine_tmst );
                    }
                    else
                    {
                        /* leave the fields empty */
                        fprintf( log_file, "," );
                    }
                }

                /* "payload (hex)", hex-encoded, bundled in 32-bit words */
                fputs( "\"", log_file );
                for( j=0; j < p->size; ++j )
                {
                    if( (j > 0) && (j%4 == 0) ) fputs( "-", log_file );
                    fprintf( log_file, "%02X", p->payload[j] );
                }
                fputs("\"\n", log_file);

                /* Flush line and count it */
                fflush( log_file );
                ++pkt_in_log;
            }

            /* Print out temperature */
            x = sx1301ar_get_temperature( 0, &u8_tempcode, &s16_temp_deg );
            if( x != 0 )
            {
                printf( "Failed to read temp\n" );
            }
            else
            {
                printf( "Temp from AD9361 = %d oC (%u)\n", s16_temp_deg, u8_tempcode );
            }

        }

        /* Check time and rotate log file if necessary */
        ++time_check;
        if( (gwc.log_rotate_period > 0) && (time_check >= 10) )
        {
            time_check = 0;
            time( &now_time );
            if( difftime( now_time, log_start_time ) > gwc.log_rotate_period )
            {
                fclose( log_file );
                printf( "INFO: log file closed, %lu packet(s) recorded\n", pkt_in_log );
                pkt_in_log = 0;
                log_start_time = now_time;
                log_file = open_log( &gwc, log_start_time );
                if( log_file == NULL )
                {
                    sx1301ar_stop( board_nb );
                    return EXIT_FAILURE;
                }
            }
        }
    }

    printf( "\nNb valid packets received: %lu CRC OK\n", nb_pkt_crc_ok );

    if( enable_gps == true )
    {
        pthread_cancel( thrid_gps ); /* don't wait for GPS thread */
    }

    /* If an exit signal was received, try to quit properly */
    if( exit_sig == 1 )
    {
        /* Stop the concentrator */
        x = sx1301ar_stop( board_nb );
        if( x != 0 )
        {
            printf( "ERROR: sx1301ar_stop failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
            return EXIT_FAILURE;
        }
        printf( "INFO: concentrator stopped\n" );

        /* Close SPI link */
        x = spi_linuxdev_close( spi1_context );
        if( x != 0 )
        {
            printf( "ERROR: closing SPI failed, returned %i\n", x );
            return EXIT_FAILURE;
        }
        printf( "INFO: SPI link closed\n" );

        /* Close GPS serial port */
        if( enable_gps == true )
        {
            /* Restore terminal attributes */
            printf( "INFO: restoring terminal attributes\n" );
            tcsetattr( fd_gps, TCSANOW, &ttyopt_restore );

            x = close( fd_gps );
            if( x != 0 )
            {
                printf( "ERROR: closing GPS serial port failed\n" );
                return EXIT_FAILURE;
            }
            printf( "INFO: GPS serial port closed\n" );
        }
    }

    /* Close log file */
    fclose( log_file );
    printf( "INFO: log file closed, %lu packet(s) recorded\n", pkt_in_log );

    printf( "+++  Exiting packet logger +++\n" );
    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void * thread_gps( const void * arg )
{
    int i, x; /* return value for sx1301ar_* functions */

    /* File descriptor for the GPS TTY (already open and configured) */
    const int fd_gps = *(int *)arg;

    char buff[256]; /* buffer to receive GPS data */
    ssize_t nb_char;
    size_t wr_idx = 0; /* pointer to end of chars in buffer */

    sx1301ar_nmea_msg_t msg_type;
    sx1301ar_ubx_msg_t ubx_type;
    int n_sat; /* GPS quality metrics */
    float hdop; /* GPS quality metrics */
    sx1301ar_coord_t coord; /* GPS 3D coordinates */

    bool trig_sync = false;
    struct timespec utc; /* UTC time */
    struct timespec timegps; /* GPS time */
    bool utc_fresh = false; /* true when UTC was updated */
    uint32_t cnt_pps; /* internal timestamp counter captured on PPS edge */
    uint32_t hs_pps; /* high speed counter captured on PPS edge */

    /* Initialize some variables before loop */
    memset( buff, 0, sizeof  buff );

    /* Loop until user quits */
    while( (quit_sig != 1) && (exit_sig != 1) )
    {
        size_t rd_idx = 0;
        size_t frame_end_idx = 0;

        /* Wait for NMEA/UBX frames (blocking non-canonical read on serial port) */
        nb_char = read( fd_gps, buff + wr_idx, LGW_GPS_MIN_MSG_SIZE );
        if( nb_char <= 0 )
        {
            printf( "WARNING: serial port read() returned value %zd\n", nb_char );
            continue;
        }
        wr_idx += (size_t)nb_char;

        /*******************************************
         * Scan buffer for UBX/NMEA sync chars and *
         * attempt to decode frame if one is found *
         *******************************************/
        while( rd_idx < wr_idx )
        {
            size_t frame_size = 0;

            /* Scan buffer for UBX sync char */
            if( buff[rd_idx] == (char)LGW_GPS_UBX_SYNC_CHAR )
            {
                /************************
                 *  Found UBX sync char *
                 ************************/

                /* Determine message type */
                ubx_type = UBX_UNDEFINED;
                x = sx1301ar_get_ubx_type( buff, (wr_idx - rd_idx), &ubx_type, &frame_size );
                if( x != 0 )
                {
                    printf( "ERROR: sx1301ar_get_ubx_type failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                    exit( EXIT_FAILURE );
                }

                /* UBX parsing tree */
                /* /!\ This code works on uBlox GPS modules, you might have to modify the code if used with another brand of GPS receivers */
                if( ubx_type == UBX_NAV_TIMEGPS)
                {
                    /* Parse NAV-TIMEGPS frame */
                    x = sx1301ar_parse_ubx_timegps( buff, frame_size, &timegps );
                    if( x != 0 )
                    {
                        printf( "ERROR: sx1301ar_parse_ubx_timegps failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                        exit( EXIT_FAILURE );
                    }
                }
            }
            else if( buff[rd_idx] == (char)LGW_GPS_NMEA_SYNC_CHAR )
            {
                /************************
                 * Found NMEA sync char *
                 ************************/

                /* scan for NMEA end marker (LF = 0x0a) */
                char * nmea_end_ptr = memchr( &buff[rd_idx], (int)0x0a, (wr_idx - rd_idx) );

                if( nmea_end_ptr )
                {
                    /* found end marker */
                    frame_size = nmea_end_ptr - &buff[rd_idx] + 1;

                    /* Determine message type */
                    msg_type = NMEA_UNDEFINED;
                    x = sx1301ar_get_nmea_type( buff, frame_size, &msg_type );
                    if( x != 0 )
                    {
                        printf( "ERROR: sx1301ar_get_nmea_type failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                        exit( EXIT_FAILURE );
                    }

                    /* NMEA parsing tree */
                    /* /!\ This code works on uBlox GPS modules, you might have to modify the code if used with another brand of GPS receivers */
                    if( msg_type == NMEA_RMC)
                    {
                        utc_fresh = false;

                        /* Parse RMC frame */
                        x = sx1301ar_parse_rmc( buff, frame_size, &utc, NULL );
                        if( x == 0 )
                        {
                            utc_fresh = true;
                            trig_sync = true; /* We need to trigger a sync each time we get an RMC sentence */
                        }
                        else if( x > 0 )
                        {
                            printf( "WARNING: parsing RMC sentence returned - ");
                            switch( x )
                            {
                                case 1:
                                    printf( "Bad checksum\n" );
                                    break;
                                case 2:
                                    printf( "Bad format\n" );
                                    break;
                                case 4:
                                    printf( "Invalid\n" );
                                    break;
                                case 8:
                                    printf( "Incomplete\n" );
                                    break;
                                default:
                                    break;
                            }
                        }
                        else if( x < 0 )
                        {
                            printf( "ERROR: sx1301ar_parse_rmc failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                            exit( EXIT_FAILURE );
                        }
                    }
                    else if( msg_type == NMEA_GGA)
                    {
                        trig_sync = false;

                        /* Parse GGA frame */
                        x = sx1301ar_parse_gga( buff, frame_size, &coord, &n_sat, &hdop );
                        if( x == 0 )
                        {
                            /* If metrics are good enough, trigger a sync */
                            /* /!\ ASSUMES GPS SENDS RMC BEFORE GGA, CHECK YOUR GPS MODULE */
                            if( (n_sat >= N_SAT_MIN) && (hdop < HDOP_MAX) && (utc_fresh == true) )
                            {
                                utc_fresh = false; /* invalidates in case GPS stops sending RMC unexpectedly */
                                trig_sync = true;
                            }
                        }
                        else if( x > 0 )
                        {
                            printf( "WARNING: parsing GGA sentence returned - ");
                            switch( x )
                            {
                                case 1:
                                    printf( "Bad checksum\n" );
                                    break;
                                case 2:
                                    printf( "Bad format\n" );
                                    break;
                                case 4:
                                    printf( "Invalid\n" );
                                    break;
                                case 8:
                                    printf( "Incomplete\n" );
                                    break;
                                default:
                                    break;
                            }
                        }
                        else if( x < 0 )
                        {
                            /* TODO: where is the best place to invalidate obsolete info ? */
                            printf( "ERROR: sx1301ar_parse_gga failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                            exit( EXIT_FAILURE );
                        }
                    }
                }
            }

            if( frame_size > 0 )
            {
                /* At this point message is a checksum verified frame
                   we're processed or ignored. Remove frame from buffer */
                rd_idx += frame_size;
                frame_end_idx = rd_idx;
            }
            else
            {
                rd_idx++;
            }
        }

        if( frame_end_idx )
        {
          /* Frames have been processed. Remove bytes to end of last processed frame */
          memcpy( buff, &buff[frame_end_idx], wr_idx - frame_end_idx );
          wr_idx -= frame_end_idx;
        }

        /* Prevent buffer overflow */
        if( (sizeof(buff) - wr_idx) < LGW_GPS_MIN_MSG_SIZE )
        {
            memcpy( buff, &buff[LGW_GPS_MIN_MSG_SIZE], wr_idx - LGW_GPS_MIN_MSG_SIZE );
            wr_idx -= LGW_GPS_MIN_MSG_SIZE;
        }

        /* Do a synchronization cycle when all needed info are available */
        if( trig_sync == true )
        {
            trig_sync = false;

            /* Get timestamp value captured on PPS edge (use mutex to protect hardware access) */
            pthread_mutex_lock( &mx_concent );
            x  = sx1301ar_get_trigcnt( SX1301AR_BOARD_MASTER, &cnt_pps );
            x |= sx1301ar_get_trighs( SX1301AR_BOARD_MASTER, &hs_pps );
            pthread_mutex_unlock( &mx_concent );
            if( x != 0 )
            {
                printf( "ERROR: sx1301ar_get_trigcnt/hs failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                return NULL;
            }

            /* Attempt sync */
            pthread_mutex_lock( &mx_tref );
            x = sx1301ar_synchronize( &time_ref, utc, timegps, cnt_pps, hs_pps, NULL );
            pthread_mutex_unlock( &mx_tref );
            if( x == 0 )
            {
                /* Set current xtal err to all boards */
                for( i = 0; i < board_nb; i++ )
                {
                    x = sx1301ar_set_xtal_err( i, time_ref );
                    if( x != 0 )
                    {
                        printf( "ERROR: failed to set XTAL error to board %d\n", i );
                    }
                }
                //printf( "Sync successful (XTAL err: %.9lf)\n", time_ref.xtal_err ); /* too verbose */
            }
            else if( x < 0 )
            {
                printf( "ERROR: sx1301ar_synchronize critically failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                exit( EXIT_FAILURE );
            }
        }
    }
    printf( "\nINFO: End of GPS thread\n" );
    return NULL;
}

static void sig_handler( int sigio )
{
    if( sigio == SIGQUIT )
    {
        quit_sig = 1;
    }
    else if( (sigio == SIGINT) || (sigio == SIGTERM) )
    {
        exit_sig = 1;
    }
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int parse_sx1301ar_configuration( JSON_Object * brd_conf_obj, uint8_t brd )
{
    int i, j, k;
    int x; /* return code for SX1301ar driver functions */
    char param_name[64]; /* used to generate variable parameter names */
    JSON_Array * conf_sx1301_array;
    JSON_Array * conf_rfchain_array;
    JSON_Object * conf_chip_obj;
    JSON_Object * conf_chan_obj;
    JSON_Object * conf_rfchain_obj;
    JSON_Value * val;
    uint8_t chan_nb;
    uint32_t tmp_u32;
    int sf_a, sf_b;
    unsigned long long ull = 0, ull2 = 0;
    const char * str;
    int str_len;

    /* Configuration structures */
    sx1301ar_board_cfg_t cfg_brd = sx1301ar_init_board_cfg( );
    sx1301ar_chip_cfg_t cfg_chp;
    sx1301ar_chan_cfg_t cfg_chn;

    /* Associate with SPI */
    cfg_brd.spi_read = &spi1_read;
    cfg_brd.spi_write = &spi1_write;

    /* Mandatory: Type of the board */
    val = json_object_get_value( brd_conf_obj, "board_type" );
    if( json_value_get_type( val ) != JSONString )
    {
        printf( "ERROR: impossible to parse mandatory parameter 'board_type'\n" );
        return -1;
    }
    str = json_value_get_string( val );
    if( strncmp( str, "MASTER", strlen(str) ) == 0 )
    {
        printf( "INFO: Master board\n" );
        cfg_brd.board_type = BRD_MASTER;
    }
    else if( strncmp( str, "SLAVE", strlen(str) ) == 0 )
    {
        printf( "INFO: Slave board\n" );
        cfg_brd.board_type = BRD_SLAVE;
    }
    else
    {
        printf( "ERROR: board type %s is not valid\n", str );
        return -1;
    }

    /* Mandatory: Rx frequency of the board (ie. RX freq of the AD9361) */
    val = json_object_get_value( brd_conf_obj, "board_rx_freq" );
    if( json_value_get_type( val ) != JSONNumber )
    {
        printf( "ERROR: impossible to parse mandatory parameter 'board_rx_freq'\n" );
        return -1;
    }
    cfg_brd.rx_freq_hz = (uint32_t)json_value_get_number( val );
    printf( "INFO: board center frequency configured at %u Hz\n", cfg_brd.rx_freq_hz );
    /* Mandatory: Rx bandwidth of the board (ie. RX bw of the AD9361) */
    val = json_object_get_value( brd_conf_obj, "board_rx_bw" );
    if( json_value_get_type( val ) != JSONNumber )
    {
        printf( "ERROR: impossible to parse mandatory parameter 'board_rx_bw'\n" );
        return -1;
    }
    cfg_brd.rx_bw_hz = (uint32_t)json_value_get_number( val );
    printf( "INFO: board RX bandwidth configured at %u Hz\n", cfg_brd.rx_bw_hz );
    /* Optional: radio mode */
    cfg_brd.full_duplex = (bool)json_object_get_boolean( brd_conf_obj, "full_duplex" );

    /* RF chains configuration */
    conf_rfchain_array = json_object_get_array( brd_conf_obj, "rf_chain_conf" );
    if( conf_rfchain_array != NULL )
    {
        rf_chain_nb = (int)json_array_get_count( conf_rfchain_array );
        for( i = 0; i < rf_chain_nb; i++ )
        {
            /* Sanity check */
            if( i >= SX1301AR_BOARD_RFCHAIN_NB )
            {
                printf( "ERROR: rf_chain %d not supported, skip it\n", i );
                break;
            }

            /* Get RF chain conf object */
            conf_rfchain_obj = json_array_get_object( conf_rfchain_array, i );

            /* Parse configuration */
            cfg_brd.rf_chain[i].rx_enable = (bool)json_object_get_boolean( conf_rfchain_obj, "rx_enable" );
            cfg_brd.rf_chain[i].tx_enable = (bool)json_object_get_boolean( conf_rfchain_obj, "tx_enable" );
            cfg_brd.rf_chain[i].rssi_offset = (float)json_object_get_number( conf_rfchain_obj, "rssi_offset" );
            cfg_brd.rf_chain[i].rssi_offset_coeff_a = (int16_t)json_object_get_number( conf_rfchain_obj, "rssi_offset_coeff_a" );
            cfg_brd.rf_chain[i].rssi_offset_coeff_b = (int16_t)json_object_get_number( conf_rfchain_obj, "rssi_offset_coeff_b" );
            printf( "INFO: rf_chain %d: RSSI offset configured at %.1f, a=%d, b=%d\n", i, cfg_brd.rf_chain[i].rssi_offset, cfg_brd.rf_chain[i].rssi_offset_coeff_a, cfg_brd.rf_chain[i].rssi_offset_coeff_b );
        }
    }
    else
    {
        printf( "WARNING: no rf_chain configured\n" );
    }

    /* Optional: FSK sync word */
    val = json_object_get_value( brd_conf_obj, "FSK_sync" );
    if( json_value_get_type( val ) == JSONString )
    {
        str = json_value_get_string( val );
        str_len = strlen( str );
        if( (str_len < 2) || (str_len > 16) || (str_len%2 != 0) )
        {
            printf( "ERROR: impossible to parse parameter 'FSK_sync'\n" );
            return -1;
        }
        i = sscanf( str, "%llx", &ull );
        if( i != 1 )
        {
            printf( "ERROR: failed to parse hex string for 'FSK_sync'\n" );
            return -1;
        }
        cfg_brd.fsk_sync_size = str_len / 2;
        cfg_brd.fsk_sync_word = (uint64_t)ull;
    }

    /* Optional: enable cohabitation setting for LoRa MAC public networks */
    cfg_brd.loramac_public = (bool)json_object_get_boolean( brd_conf_obj, "loramac_public" );

    /* Getting AES-128 decryption key */
    val = json_object_get_value( brd_conf_obj, "aes_key" );
    if( json_value_get_type( val ) == JSONString )
    {
        str = json_value_get_string( val );
        str_len = strlen( str );
        if( str_len != 32 ) /* 16 bytes => 32 characters */
        {
            printf( "ERROR: wrong size for AES-128 key\n" );
        }
        else
        {
            i = sscanf( json_value_get_string( val ), "%16llx%16llx", &ull, &ull2 );
            if( i != 2 )
            {
                printf( "ERROR: failed to parse hex string for AES-128 key\n" );
            }
            else
            {
                for( i = 0; i < 8; i++ )
                {
                    cfg_brd.aes_key[i]   = (uint8_t)((ull  >> (56 - i*8)) & 0xFF);
                    cfg_brd.aes_key[i+8] = (uint8_t)((ull2 >> (56 - i*8)) & 0xFF);
                }
                printf( "INFO: gateway AES-128 key is configured to %016llX%016llX\n", ull, ull2 );
            }
        }
    }
    else
    {
        printf( "ERROR: wrong format for AES-128 key\n" );
    }

    /* Get room temperature used as Tref reference */
    cfg_brd.room_temp_ref = (int8_t)json_object_get_number( brd_conf_obj, "calibration_temperature_celsius_room" );
    printf( "INFO: Calibration room temperature (Tref) set to %u oC\n", cfg_brd.room_temp_ref );

    /* Get radio temperature sensor reference value (for Tref°C) */
    cfg_brd.ad9361_temp_ref = (uint8_t)json_object_get_number( brd_conf_obj, "calibration_temperature_code_ad9361" );
    printf( "INFO: Calibration radio temperature code at [Tref] set to %u\n", cfg_brd.ad9361_temp_ref );

    /* Get DSP status report interval */
    cfg_brd.dsp_stat_interval = (uint8_t)json_object_get_number( brd_conf_obj, "dsp_stat_interval" );

    /* Get the number of dsp on the board */
    val = json_object_get_value( brd_conf_obj, "nb_dsp" );
    if( json_value_get_type( val ) != JSONNumber )
    {
        printf( "ERROR: impossible to parse mandatory parameter 'nb_dsp'\n" );
        return -1;
    }
    cfg_brd.nb_dsp = (uint32_t)json_value_get_number( val );
    printf( "INFO: %u dsp on the board\n", cfg_brd.nb_dsp );

    /* Get the number of chip on the board */
    conf_sx1301_array = json_object_get_array( brd_conf_obj, "SX1301_conf" );
    if( conf_sx1301_array != NULL )
    {
        cfg_brd.nb_chip = (uint8_t)json_array_get_count( conf_sx1301_array );
    }

    /* Commit board configuration */
    x = sx1301ar_conf_board( brd, &cfg_brd );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_conf_board failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return -1;
    }

    /* SX1301 chip configuration */
    if( conf_sx1301_array != NULL )
    {
        /* Chips configuration loop */
        for( i = 0; i < (int)json_array_get_count( conf_sx1301_array ); ++i )
        {
            /* Sanity check */
            if( i >= SX1301AR_BOARD_CHIPS_NB)
            {
                printf( "ERROR: SX1301 chip %d not supported, skip it\n", i );
                break;
            }

            /* (re)initialize chip configuration structure */
            cfg_chp = sx1301ar_init_chip_cfg( );

            /* Point to chip configuration */
            conf_chip_obj = json_array_get_object( conf_sx1301_array, i );

            cfg_chp.enable = (bool)json_object_get_boolean( conf_chip_obj, "chip_enable" );
            if( cfg_chp.enable == false )
            {
                printf( "INFO: chip %i disabled\n", i );
                x = sx1301ar_conf_chip( brd, i, &cfg_chp );
                if( x != 0 )
                {
                    printf( "ERROR: sx1301ar_conf_chip failed for chip %i; %s\n", i, sx1301ar_err_message( sx1301ar_errno ) );
                    return -1;
                }
                /* Skip to next chip */
                continue;
            }

            /* Configure enbaled chip */
            printf( "INFO: chip %i enabled\n", i );
            /* Mandatory: center frequency of the chip (ie. RX freq of each SX1301) */
            val = json_object_get_value( conf_chip_obj, "chip_center_freq" );
            if( json_value_get_type( val ) != JSONNumber )
            {
                printf( "ERROR: failed to parse mandatory parameter chip_center_freq for chip %i\n", i );
                return -1;
            }
            cfg_chp.freq_hz = (uint32_t)json_value_get_number( val );

            /* Rx RF chain select */
            val = json_object_get_value( conf_chip_obj, "chip_rf_chain" );
            if( json_value_get_type( val ) != JSONNumber )
            {
                printf( "ERROR: failed to parse mandatory parameter chip_rf_chain for chip %i\n", i );
                return -1;
            }
            cfg_chp.rf_chain = (uint8_t)json_value_get_number( val );

            /* Commit chip configuration */
            printf( "INFO: chip %i center frequency configured at %u Hz\n", i, cfg_chp.freq_hz );
            x = sx1301ar_conf_chip( brd, i, &cfg_chp );
            if( x != 0 )
            {
                printf( "ERROR: sx1301ar_conf_chip failed for chip %i; %s\n", i, sx1301ar_err_message( sx1301ar_errno ) );
                return -1;
            }

            /* Channels configuration loop */
            for( j = 0; j < SX1301AR_CHIP_CHAN_NB; ++j )
            {
                chan_nb = (i << 4) + j; /* generate board-level channel number */
                cfg_chn = sx1301ar_init_chan_cfg( ); /* (re)initialize channel configuration structure */

                /* Point to channel configuration */
                if( j < SX1301AR_CHIP_MULTI_NB )
                {
                    sprintf( param_name, "chan_multiSF_%i", j );
                    conf_chan_obj = json_object_get_object( conf_chip_obj, param_name );
                }
                else if( j == SX1301AR_CHIP_LSA_IDX )
                {
                    conf_chan_obj = json_object_get_object( conf_chip_obj, "chan_LoRa_std" );
                }
                else if( j == SX1301AR_CHIP_FSK_IDX )
                {
                    conf_chan_obj = json_object_get_object( conf_chip_obj, "chan_FSK" );
                }

                /* Channel disabled if configuration is absent */
                if( conf_chan_obj == NULL )
                {
                    cfg_chn.enable = false; /* should already be default value */
                    printf( "INFO: channel 0x%02X disabled\n", chan_nb );
                    x = sx1301ar_conf_chan( brd, chan_nb, &cfg_chn );
                    if( x != 0 )
                    {
                        printf( "ERROR: sx1301ar_conf_chan failed for channel 0x%02X; %s\n", chan_nb, sx1301ar_err_message( sx1301ar_errno ) );
                        return -1;
                    }
                    continue;
                }
                else
                {
                    cfg_chn.enable = true;
                }

                /* Mandatory: center frequency of the channel */
                val = json_object_get_value( conf_chan_obj, "chan_rx_freq" );
                if( json_value_get_type( val ) != JSONNumber )
                {
                    printf( "ERROR: failed to parse mandatory parameter chan_rx_freq for chip 0x%02X\n", chan_nb );
                    return -1;
                }
                cfg_chn.freq_hz = (uint32_t)json_value_get_number( val );

                /* Parse parameters specific to each type of channel */
                if( j < SX1301AR_CHIP_MULTI_NB )
                {
                    cfg_chn.bandwidth = BW_125K; /* not really necessary */

                    /* LoRa spreading factor or spreading factor range (multi-SF modem) */
                    val = json_object_get_value( conf_chan_obj, "spread_factor" );
                    if( json_value_get_type( val ) == JSONString )
                    {
                        k = sscanf( json_value_get_string( val ), "%i-%i", &sf_a, &sf_b );
                        if( k == 2 )
                        {
                            cfg_chn.modrate = sx1301ar_sf_range_nb2enum( sf_a, sf_b );
                        }
                        else if( k == 1 )
                        {
                            cfg_chn.modrate = sx1301ar_sf_nb2enum( sf_a );
                        }
                        else
                        {
                            cfg_chn.modrate = MR_UNDEFINED;
                        }
                    }
                    else if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_chn.modrate = sx1301ar_sf_nb2enum( (int)json_value_get_number( val ) );
                    }
                    else
                    {
                        cfg_chn.modrate = MR_UNDEFINED;
                    }
                }
                else if( j == SX1301AR_CHIP_LSA_IDX )
                {
                    /* Bandwidth */
                    tmp_u32 = (uint32_t)json_object_get_number( conf_chan_obj, "bandwidth" );
                    cfg_chn.bandwidth = sx1301ar_bw_nb2enum( tmp_u32 );

                    /* SF */
                    tmp_u32 = (uint32_t)json_object_get_number( conf_chan_obj, "spread_factor" );
                    cfg_chn.modrate = sx1301ar_sf_nb2enum( tmp_u32 );
                }
                else if( j == SX1301AR_CHIP_FSK_IDX )
                {
                    /* Bandwidth */
                    tmp_u32 = (uint32_t)json_object_get_number( conf_chan_obj, "bandwidth" );
                    cfg_chn.bandwidth = sx1301ar_bw_nb2enum( tmp_u32 );

                    /* Bit rate */
                    tmp_u32 = (uint32_t)json_object_get_number( conf_chan_obj, "bit_rate" );
                    cfg_chn.modrate = tmp_u32;
                }

                /* Commit channel configuration */
                if( j < SX1301AR_CHIP_MULTI_NB )
                {
                    printf( "INFO: LoRa multi-SF channel 0x%02x configured at %u Hz, BW 125kHz, SF %i to %i\n", chan_nb, cfg_chn.freq_hz, sx1301ar_sf_min_enum2nb( cfg_chn.modrate ), sx1301ar_sf_max_enum2nb( cfg_chn.modrate ) );
                }
                else if( j == SX1301AR_CHIP_LSA_IDX )
                {
                    printf( "INFO: LoRa stand-alone channel 0x%02x configured at %u Hz, BW %li Hz, SF %i\n", chan_nb, cfg_chn.freq_hz, sx1301ar_bw_enum2nb( cfg_chn.bandwidth ), sx1301ar_sf_enum2nb( cfg_chn.modrate ) );
                }
                else if( j == SX1301AR_CHIP_FSK_IDX )
                {
                    printf( "INFO: FSK channel 0x%02x configured at %u Hz, BW %li Hz, bit rate %li\n", chan_nb, cfg_chn.freq_hz, sx1301ar_bw_enum2nb( cfg_chn.bandwidth ), (long)cfg_chn.modrate );
                }
                x = sx1301ar_conf_chan( brd, chan_nb, &cfg_chn );
                if( x != 0 )
                {
                    printf( "ERROR: sx1301ar_conf_chan failed for channel 0x%02X; %s\n", chan_nb, sx1301ar_err_message( sx1301ar_errno ) );
                    return -1;
                }
            }
        }
    }

    return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int parse_gateway_configuration( JSON_Object * gw_conf_obj, gw_conf_t * gwc )
{
    JSON_Value * val;
    unsigned long long ull = 0;

    /* Getting network parameters (only those necessary for the packet logger) */
    val = json_object_get_value( gw_conf_obj, "gateway_ID" );
    if( json_value_get_type( val ) == JSONString )
    {
        sscanf( json_value_get_string( val ), "%llx", &ull );
        gwc->gw_id = ull;
    }
    else
    {
        gwc->gw_id = 0;
    }
    printf( "INFO: gateway MAC address is configured to %016llX\n", ull );

    /* Optional: log rotation parameter */
    val = json_object_get_value( gw_conf_obj, "log_rotate_period" );
    if( json_value_get_type( val ) == JSONNumber )
    {
        gwc->log_rotate_period = (int)json_value_get_number( val );
    }

    return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static FILE * open_log( const gw_conf_t * gwc, time_t log_time )
{
    int i, j;
    char iso_date[20];
    char log_file_name[64];
    FILE * log_file;

    strftime( iso_date, ARRAY_SIZE( iso_date ), "%Y%m%dT%H%M%SZ", gmtime( &log_time ) ); /* format yyyymmddThhmmssZ */
    sprintf( log_file_name, "pktlog_%s_%s.csv", gwc->gw_id_str, iso_date );

    log_file = fopen( log_file_name, "a" ); /* create log file, append if file already exist */
    if( log_file == NULL )
    {
        printf( "ERROR: impossible to create log file %s\n", log_file_name );
        return NULL;
    }

    i = fprintf( log_file, "\"sys time (UTC)\",\"GPS timestamp\",\"board\",\"counter timestamp\",\"frequency\"," );
    if( i < 0 )
    {
        printf( "ERROR: impossible to write to log file %s\n", log_file_name );
        return NULL;
    }

    for( j = 0; j < rf_chain_nb; j++ )
    {
        i = fprintf( log_file, "\"channel nb (ant%d)\",\"RSSI (ant%d)\",", j, j );
        if( i < 0 )
        {
            printf( "ERROR: impossible to write to log file %s\n", log_file_name );
            return NULL;
        }
    }

    i = fprintf( log_file, "\"status\",\"size\",\"modulation\",\"BW (kHz)\",\"SF\",\"data rate\",\"CR\"," );
    if( i < 0 )
    {
        printf( "ERROR: impossible to write to log file %s\n", log_file_name );
        return NULL;
    }

    for( j = 0; j < rf_chain_nb; j++ )
    {
        i = fprintf( log_file, "\"SNR (ant%d)\",", j );
        if( i < 0 )
        {
            printf( "ERROR: impossible to write to log file %s\n", log_file_name );
            return NULL;
        }
    }

    for( j = 0; j < rf_chain_nb; j++ )
    {
        i = fprintf( log_file, "\"ftime (ant%d)\",", j );
        if( i < 0 )
        {
            printf( "ERROR: impossible to write to log file %s\n", log_file_name );
            return NULL;
        }
    }

    i = fprintf( log_file, "\"payload (hex)\"\n" );
    if( i < 0 )
    {
        printf( "ERROR: impossible to write to log file %s\n", log_file_name );
        return NULL;
    }

    /* Create "pktlog.csv" symlink to simplify user life */
    unlink( "pktlog.csv" );
    i = symlink( log_file_name, "pktlog.csv" );
    if( i < 0 )
    {
        printf( "ERROR: impossible to create symlink to log file %s\n", log_file_name );
        return NULL;
    }

    printf( "INFO: Now writing to log file %s\n", log_file_name );
    return log_file;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void usage( void )
{
    printf( "~~~ Library version string~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " %s\n", sx1301ar_version_info( 0, NULL, NULL ) );
    printf( "~~~ Available options ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " -h  print this help\n" );
    printf( " -d <path>  use Linux SPI device driver\n" );
    printf( "  => default path: " LINUXDEV_PATH_DEFAULT "\n" );
    printf( " -g <optional path>  use GPS receiver for synchronization\n" );
    printf( "            => default path: " GPSTTY_PATH_DEFAULT "\n" );
    printf( "            warning: NO SPACE between -g and gps path!!\n" );
    printf( " -c <filename>  use config file other than 'config.json'\n" );
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */
