/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    LoRa Gateway v2 packet sender program

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Michael Coracin
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <signal.h>     /* sigaction */
#include <string.h>     /* strcmp */
#include <getopt.h>     /* getopt_long */

#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"
#include "sx1301ar_radio.h"

#include "spi_linuxdev.h"

#include "parson.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

#define AD9361_ATTEN_MAX            359 /* cf AD9361 datasheet */
#define AD9361_VREF_MAX             3
#define AD9361_AUXDAC_MAX           1023

#define DEFAULT_RADIO_RX_FREQ   866500000   /* RX center frequency of the radio */
#define DEFAULT_RADIO_RX_BW     7000000     /* RX bandwidth of the radio */
#define DEFAULT_START_FREQ      863000000   /* frequency of first channel, Hz */
#define DEFAULT_CHAN_NB         1           /* nb on channels to test */
#define DEFAULT_FREQ_OFFSET     200000      /* frequency offset between channels, Hz */
#define DEFAULT_RF_POWER        27          /* TX RF power, dBm */

#define DEFAULT_LORA_BW         125     /* LoRa modulation bandwidth, kHz */
#define DEFAULT_LORA_SF         7       /* LoRa SF */
#define DEFAULT_LORA_CR         1       /* LoRa CR */
#define DEFAULT_LORA_INVERT     false   /* LoRa modulation polarity control */
#define DEFAULT_PAYLOAD_SIZE    32      /* payload size, bytes */

#define LBT_RSSI_SHIFT              142     /* to convert RSSI target in dBm to register value */
#define LBT_RSSI_TARGET_MIN         -105    /* dBm */
#define LBT_RSSI_TARGET_MAX         -42     /* dBm */
#define LBT_NUMBER_OF_CHANNELS_MAX  16
#define LBT_CHANNEL_FREQ_STEP       200E3   /* Hz */
#define LBT_SCAN_TIME1              128     /* µs */
#define LBT_SCAN_TIME2              5000    /* µs */
#define LBT_TRANSMIT_TIME           400     /* ms */

#define DEFAULT_BR_KBPS             50
#define DEFAULT_FDEV_KHZ            25

#define DEFAULT_LINUXDEV_PATH   "/dev/spidev0.0"

#define JSON_SX1301AR_CONF_OBJ  "SX1301_array_conf"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* Signal handling variables */
static int exit_sig = 0; /* 1 -> application terminates cleanly (shut down hardware, close open files, etc) */
static int quit_sig = 0; /* 1 -> application terminates without shutting down the hardware */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* TX gain lut */
static sx1301ar_tx_gain_lut_t txgain_lut = {
    .size = 4,
    .lut[0] = {
        .fpga_dig_gain = 5,
        .ad9361_gain = {
            .atten = 30,
            .auxdac_vref = 3,
            .auxdac_word = 770
        },
        .ad9361_tcomp.coeff_a = 0,
        .ad9361_tcomp.coeff_b = 0,
        .rf_power = 27
    },
    .lut[1] = {
        .fpga_dig_gain = 5,
        .ad9361_gain = {
            .atten = 61,
            .auxdac_vref = 3,
            .auxdac_word = 770
        },
        .ad9361_tcomp.coeff_a = 0,
        .ad9361_tcomp.coeff_b = 0,
        .rf_power = 20
    },
    .lut[2] = {
        .fpga_dig_gain = 5,
        .ad9361_gain = {
            .atten = 87,
            .auxdac_vref = 3,
            .auxdac_word = 770
        },
        .ad9361_tcomp.coeff_a = 0,
        .ad9361_tcomp.coeff_b = 0,
        .rf_power = 14
    },
    .lut[3] = {
        .fpga_dig_gain = 5,
        .ad9361_gain = {
            .atten = 135,
            .auxdac_vref = 3,
            .auxdac_word = 770
        },
        .ad9361_tcomp.coeff_a = 0,
        .ad9361_tcomp.coeff_b = 0,
        .rf_power = 2
    }
};

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

static void sig_handler( int sigio );

static void usage( void );

static int ad9361_set_atten_offset( uint8_t brd, uint8_t rf_chain, int16_t atten_offset);

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );
int extract_rfchain_configuration( JSON_Array * conf_rfchain_array, sx1301ar_board_cfg_t *cfg_brd );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
    static struct sigaction sigact; /* SIGQUIT&SIGINT&SIGTERM signal handling */

    int i, j; /* loop and temporary variables */
    int x; /* return code for SX1301ar driver functions */
    unsigned cnt; /* cycles counter */
    sx1301ar_tstat_t stat;
    uint32_t tx_timeout_us;
    bool tx_timeout;

    /* SPI interfaces */
    const char spi1_path_default[] = DEFAULT_LINUXDEV_PATH;
    const char * spi1_path = spi1_path_default;

    /* Configuration structures */
    sx1301ar_board_cfg_t cfg_brd = sx1301ar_init_board_cfg( );
    sx1301ar_lbt_cfg_t cfg_lbt = sx1301ar_init_lbt_cfg( );
    bool loramac_public = false;

    /* FSK options */
    int fskmode = 0;
    float fskrate = DEFAULT_BR_KBPS;
    unsigned fskdev = DEFAULT_FDEV_KHZ;

    /* Parameter parsing */
    int option_index = 0;
    static struct option long_options[] = {
        {"rfreq", 1, 0, 0},
        {"rbw", 1, 0, 0},
        {"lbt-freq", 1, 0, 0},
        {"lbt-sctm", 1, 0, 0},
        {"lbt-trtm", 1, 0, 0},
        {"lbt-rssi", 1, 0, 0},
        {"lbt-nbch", 1, 0, 0},
        {"lbt-rssi-shift", 1, 0, 0},
        {"atten", 1, 0, 0},
        {"dac", 1, 0, 0},
        {"vref", 1, 0, 0},
        {"pad", 1, 0, 0},
        {"payload", 1, 0, 0},
        {"dyn-atten", 1, 0, 0},
        {"fsk", 0, 0, 0},
        {"fskdev", 1, 0, 0},
        {"fskrate", 1, 0, 0},
        {0, 0, 0, 0}
    };
    bool parse_err = false;
    double arg_freq = 0, arg_offset = 0;
    double arg_lf = 0;
    unsigned arg_u = 0;
    int arg_i = 0;

    /* Radio settings */
    uint32_t rx_freq_hz = DEFAULT_RADIO_RX_FREQ;
    uint32_t rx_bw_hz = DEFAULT_RADIO_RX_BW;

    /* TX settings (default parameters) */
    uint32_t start_freq = DEFAULT_START_FREQ;
    uint8_t channels = DEFAULT_CHAN_NB;
    uint32_t chan_offset = DEFAULT_FREQ_OFFSET;
    int8_t rf_power = DEFAULT_RF_POWER;
    uint16_t bandwidth_khz = DEFAULT_LORA_BW;
    uint8_t spread_factor = DEFAULT_LORA_SF;
    uint8_t coding_rate = DEFAULT_LORA_CR;
    uint16_t preamb_length = 0; /* 0 = use default HAL value */
    uint8_t pl_size = DEFAULT_PAYLOAD_SIZE;
    uint32_t pkt_delay_ms = 0;
    int32_t pkt_repeat = -1; /* -1 repeat indefinitely */
    bool invert_enable = DEFAULT_LORA_INVERT;
    uint8_t rf_chain = 0;
    uint32_t sx1301_count_us;

    /* LBT settings */
    bool lbt_enable = false;
    uint32_t lbt_f_target = 0;
    uint32_t lbt_scan_time = 5000;
    uint32_t lbt_transmit_time = 4000;
    int8_t lbt_rssi_target_dBm = -80;
    uint8_t lbt_rssi_shift = LBT_RSSI_SHIFT;
    uint8_t lbt_nb_channel = 8;

    /* Dynamic atten */
    uint32_t dynatten_timing = 0;
    int dynatten_offset = 12;

    /* Paayload values */
    int padvalue = -1;
    uint8_t payload[256];

    /* TX error rate */
    uint32_t nb_sent = 0;
    uint32_t nb_error = 0;

    /* JSON variables */
    JSON_Value * root_val;
    JSON_Array * conf_sx1301_array = NULL;
    JSON_Array * conf_rfchain_array = NULL;
    JSON_Object * root = NULL;
    JSON_Object * brd_conf_obj = NULL;

    char userchar[10];
    fd_set readfds;
    struct timeval tv;
    int fd_stdin;

    /* Parse command line options */
    while( (i = getopt_long( argc, argv, "hd:w:f:b:s:c:p:r:z:t:x:ima:j:", long_options, &option_index )) != -1 )
    {
        switch( i )
        {
        case 'h':
            usage( );
            return EXIT_SUCCESS;
            break;

        case 'd': /* -d <path>  use Linux SPI device driver */
            if( optarg != NULL )
            {
                spi1_path = optarg;
            }
            break;

        case 'f':
            /*
            -f <float>  target frequency in MHz
            -f <>:<unsigned>  previous + number of channels to jump
            -f <>:<>:<float>  previous + explicit offset in MHz between channels
            */
            j = sscanf( optarg, "%lf:%u:%lf", &arg_freq, &arg_u, &arg_offset );
            printf( "debug: %i, %lf, %u, %lf\n", j, arg_freq, arg_u, arg_offset);
            switch( j )
            {
            case 3:
                if( (arg_offset < 0.01) || (arg_offset > 30.0) )
                {
                    parse_err = true;
                }
                else
                {
                    chan_offset = (uint32_t)((arg_offset * 1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
                }
                /* Fall through */
            case 2:
                if( (arg_u == 0) || (arg_u > 100) )
                {
                    parse_err = true;
                }
                else
                {
                    channels = arg_u;
                }
                /* Fall through */
            case 1:
                if( (arg_freq < 30.0) || (arg_freq > 3000.0) )
                {
                    parse_err = true;
                }
                else
                {
                    start_freq = (uint32_t)((arg_freq * 1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
                }
                break;

            default:
                parse_err = true;

            }
            if( parse_err )
            {
                printf( "ERROR: argument parsing of -f argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            break;

        case 'b': /* -b <uint>  LoRa modulation bandwidth */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || ((arg_u != 125) && (arg_u != 250) && (arg_u != 500)) )
            {
                printf( "ERROR: argument parsing of -b argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                bandwidth_khz = (uint16_t)arg_u;
            }
            break;

        case 's': /* -s <uint>  LoRa Spreading Factor */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u < 7) || (arg_u > 12) )
            {
                printf( "ERROR: argument parsing of -s argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                spread_factor = (uint8_t)arg_u;
            }
            break;

        case 'c': /* -c <uint>  LoRa Coding Rate */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u < 1) || (arg_u > 4) )
            {
                printf( "ERROR: argument parsing of -c argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                coding_rate = (uint8_t)arg_u;
            }
            break;

        case 'p': /* -p <int>  RF power (dBm) */
            j = sscanf( optarg, "%i", &arg_i );
            if( (j != 1) || (arg_i < -60) || (arg_i > 60) )
            {
                printf( "ERROR: argument parsing of -p argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                rf_power = (int8_t)arg_i;
            }
            break;

        case 'r': /* -r <uint>  preamble length (symbols) */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u < 6) )
            {
                printf( "ERROR: argument parsing of -r argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                preamb_length = (uint16_t)arg_u;
            }
            break;

        case 'z': /* -z <uint>  payload length (bytes) */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u < 1) || (arg_u > 255) )
            {
                printf( "ERROR: argument parsing of -z argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                pl_size = (uint8_t)arg_u;
            }
            break;

        case 't': /* -t <uint>  pause between packets (ms) */
            j = sscanf( optarg, "%u", &arg_u );
            if( j != 1 )
            {
                printf( "ERROR: argument parsing of -t argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                pkt_delay_ms = (uint32_t)arg_u;
            }
            break;

        case 'x': /* -x <int>  nb of times the sequence is repeated */
            j = sscanf( optarg, "%i", &arg_i );
            if( (j != 1) || (arg_i < -1) )
            {
                printf( "ERROR: argument parsing of -x argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                pkt_repeat = (int32_t)arg_i;
            }
            break;

        case 'i': /* -i  send packet using inverted modulation polarity */
            invert_enable = true;
            break;

        case 'm': /* -m  configure the board in LoRa MAC *public network* mode */
            loramac_public = true;
            break;

        case 'a': /* -a <uint>  RF chain */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u > 1) )
            {
                printf( "ERROR: argument parsing of -a argument\n" );
                usage( );
                return EXIT_FAILURE;
            }
            else
            {
                rf_chain = (uint8_t)arg_u;
            }
            break;

        case 'j': /* -j  Get Calibration (Lut table) from a JSON file */
            printf( "INFO: found calibration file %s\n", optarg );
            root_val = json_parse_file_with_comments( optarg );
            root = json_value_get_object( root_val );
            if( root == NULL )
            {
                printf( "ERROR: %s is not a valid JSON file\n", optarg );
                usage();
                return EXIT_FAILURE;
            }
            conf_sx1301_array = json_object_get_array( root, JSON_SX1301AR_CONF_OBJ );
            if( conf_sx1301_array != NULL )
            {
                brd_conf_obj = json_array_get_object( conf_sx1301_array, 0 );
                if( brd_conf_obj == NULL )
                {
                    printf( "ERROR: %s array does not include item no%u\n", JSON_SX1301AR_CONF_OBJ, 0 );
                    usage();
                    return EXIT_FAILURE;
                }
            }
            else
            {
                printf( "ERROR: configuration file does not contain a JSON object named %s\n", JSON_SX1301AR_CONF_OBJ );
                return EXIT_FAILURE;
            }
            break;

        case 0:
            if( strcmp( long_options[option_index].name, "rfreq" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_lf );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --rfreq argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    rx_freq_hz = (uint32_t)( arg_lf * 1e6 );
                }
            }
            else if( strcmp( long_options[option_index].name, "rbw" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_lf );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --rbw argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    rx_bw_hz = (uint32_t)( arg_lf * 1e6 );
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-freq" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_freq );
                if( (i != 1) || (arg_freq < 30.0) || (arg_freq > 3000.0) )
                {
                    printf( "ERROR: argument parsing of --lbt-freq argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_f_target = (uint32_t)((arg_freq*1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
                    lbt_enable = true;
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-rssi" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_i );
                if( (i != 1) || (arg_i < LBT_RSSI_TARGET_MIN) || (arg_i > LBT_RSSI_TARGET_MAX) )
                {
                    printf( "ERROR: argument parsing of --lbt-rssi argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_rssi_target_dBm = (int8_t)arg_i;
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-rssi-shift" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_i );
                if( (i != 1) || (arg_i < 0) || (arg_i > 255) )
                {
                    printf( "ERROR: argument parsing of --lbt-rssi-shift argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_rssi_shift = (int8_t)arg_i;
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-sctm" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_i );
                if( (i != 1) || ((arg_i != LBT_SCAN_TIME1) && (arg_i != LBT_SCAN_TIME2)) )
                {
                    printf( "ERROR: argument parsing of --lbt-sctm argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_scan_time = (uint16_t)arg_i;
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-trtm" ) == 0 )
            {
                i = sscanf( optarg, "%u", &arg_u );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --lbt-trtm argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_transmit_time = (uint16_t)arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "lbt-nbch" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_i );
                if( (i != 1) || ((arg_i < 1) || (arg_i > LBT_NUMBER_OF_CHANNELS_MAX)) )
                {
                    printf( "ERROR: argument parsing of --lbt-nbch argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    lbt_nb_channel = (uint8_t)arg_i;
                }
            }
	    else if( strcmp( long_options[option_index].name, "atten" ) == 0 )
            {
                i = sscanf( optarg, "%u", &arg_u );
                if( (i != 1) || (arg_u > AD9361_ATTEN_MAX) )
                {
                    printf( "ERROR: argument parsing of --atten argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    txgain_lut.lut[0].ad9361_gain.atten = (uint16_t) arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "vref" ) == 0 )
            {
                i = sscanf( optarg, "%u", &arg_u );
                if( (i != 1) || (arg_u > AD9361_VREF_MAX) )
                {
                    printf( "ERROR: argument parsing of --vref argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    txgain_lut.lut[0].ad9361_gain.auxdac_vref = (uint8_t) arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "dac" ) == 0 )
            {
                i = sscanf( optarg, "%u", &arg_u );
                if( (i != 1) || (arg_u > AD9361_AUXDAC_MAX) )
                {
                    printf( "ERROR: argument parsing of --dac argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    txgain_lut.lut[0].ad9361_gain.auxdac_word = (uint16_t) arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "dyn-atten" ) == 0 )
             {
                 i = sscanf( optarg, "%u:%i", &arg_u, &arg_i );
                 if(i < 1)
                 {
                     printf( "ERROR: argument parsing of --dyn-atten argument. Use -h to print help\n" );
                     return EXIT_FAILURE;
                 }
                 else
                 {
                     dynatten_timing = arg_u;
                     if(i == 2) dynatten_offset = arg_i;
                 }
             }
            else if( strcmp( long_options[option_index].name, "pad" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_u );
                if(i != 1)
                {
                    printf( "ERROR: argument parsing of --pad argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    padvalue = arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "payload" ) == 0 )
            {
                i = 0;
                do {
                    x = sscanf( optarg+(i*2), "%02hhX", &payload[i] );
                    i++;
                } while( x > 0 );

                if(i > 0)
                {
                    pl_size = i;
                }
                else
                {
                    printf( "ERROR: argument parsing of --pad argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
            }
            else if( strcmp( long_options[option_index].name, "fsk" ) == 0 )
            {
                fskmode = 1;
            }
            else if( strcmp( long_options[option_index].name, "fskdev" ) == 0 )
            {
                i = sscanf( optarg, "%i", &arg_u );
                if(i != 1)
                {
                    printf( "ERROR: argument parsing of --fskdev argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    fskmode = 1;
                    fskdev = arg_u;
                }
            }
            else if( strcmp( long_options[option_index].name, "fskrate" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_lf );
                if(i != 1)
                {
                    printf( "ERROR: argument parsing of --fskrate argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    fskmode = 1;
                    fskrate = arg_lf;
                }
            }
            break;

        default:
            printf( "ERROR: argument parsing options, use -h option for help\n" );
            usage( );
            return EXIT_FAILURE;
        }
    }

    /* Configure signal handling */
    sigemptyset( &sigact.sa_mask );
    sigact.sa_flags = 0;
    sigact.sa_handler = sig_handler;
    sigaction( SIGQUIT, &sigact, NULL );
    sigaction( SIGINT, &sigact, NULL );
    sigaction( SIGTERM, &sigact, NULL );

    /* Start message */
    printf( "+++ Start of packet sender for 'SX1301 array' concentrator +++\n" );
    printf( "SX1301 array library version string: %s\n", sx1301ar_version_info( 0, NULL, NULL ) );

    /* Open SPI link */
    x = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( x != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i\n", x );
        return EXIT_FAILURE;
    }

    /* Configure SPI devices to master/slave modes (Needed for ATMEL CPU only)
         SPI 0: HOST <-> FPGA
         SPI 1: HOST/DSP <-> Flash
    */
    x = spi_set_mode( 0, SPI_MODE_MASTER );
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", x );
        return EXIT_FAILURE;
    }
    x = spi_set_mode( 1, SPI_MODE_SLAVE ); /* So that DSP can access the flash for booting */
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", x );
        return EXIT_FAILURE;
    }

    /* Configure the board */
    cfg_brd.spi_read = &spi1_read;
    cfg_brd.spi_write = &spi1_write;
    cfg_brd.board_type = BRD_MASTER;
    cfg_brd.nb_chip = 2; /* need both free running and GPS timer chip */

    /* Cannot set the radio in TX-only, so pick whatever RX frequency in the band */
    cfg_brd.rx_freq_hz = rx_freq_hz;
    cfg_brd.rx_bw_hz = rx_bw_hz;
    cfg_brd.full_duplex = false;
    cfg_brd.rf_chain[rf_chain].rx_enable = true;
    cfg_brd.rf_chain[rf_chain].tx_enable = true;
    cfg_brd.rf_chain[0].tx_lut = txgain_lut;
    cfg_brd.rf_chain[rf_chain].tx_lut = txgain_lut;
    cfg_brd.loramac_public = loramac_public;
    cfg_brd.fsk_sync_size = 3;
    cfg_brd.fsk_sync_word = 0xC194C1; /* C194C1 */

    if(brd_conf_obj != NULL)
    {
        /* RF chains configuration */
        conf_rfchain_array = json_object_get_array( brd_conf_obj, "rf_chain_conf" );
        x = extract_rfchain_configuration( conf_rfchain_array, &cfg_brd );
    }

    x = sx1301ar_conf_board( 0, &cfg_brd );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_conf_board failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    if( lbt_enable == true )
    {
        cfg_lbt.enable = true;
        cfg_lbt.rssi_target = lbt_rssi_target_dBm;
        cfg_lbt.rssi_shift = lbt_rssi_shift;
        cfg_lbt.nb_channel = lbt_nb_channel;
        for( i = 0; i < cfg_lbt.nb_channel; i++ )
        {
            cfg_lbt.channels[i].freq_hz = lbt_f_target + i * 200E3;
            cfg_lbt.channels[i].scan_time_us = lbt_scan_time;
            cfg_lbt.channels[i].transmit_time_ms = lbt_transmit_time;
        }
        x = sx1301ar_conf_lbt( 0, &cfg_lbt );
        if( x != 0 )
        {
            printf( "ERROR: sx1301ar_conf_lbt failed; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
            return EXIT_FAILURE;
        }
    }

    /* Start the concentrator */
    x = sx1301ar_start( 1 );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_start failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( 1 );
        return EXIT_FAILURE;
    }
    printf( "INFO: concentrator started\n" );

    if( channels == 1 )
    {
        printf( "NOTE: emitting on a single channel @%.3lfMHz\n", (double)start_freq * 1e-6 );
    }
    else
    {
        printf( "NOTE: emitting on %u channel starting at @%.3lfMHz with %.3lfMHz offset\n", channels, (double)start_freq * 1e-6, (double)chan_offset * 1e-6 );
    }

    /* Setting fixed TX parameters */
    sx1301ar_tx_pkt_t ptx = sx1301ar_init_tx_pkt( );
    if( lbt_enable == true )
    {
        /* Immediate TX is not allowed when LBT is enabled */
        ptx.tx_mode = TX_TIMESTAMPED;
        /* Wait a moment for LBT counters to have started, otherwise first TXs will be blocked */
        sx1301ar_wait_ms( 1000 );
    }
    else
    {
        ptx.tx_mode = TX_IMMEDIATE;
    }
    ptx.rf_chain = rf_chain;
    ptx.rf_power = rf_power;
    ptx.freq_hz = start_freq;
    ptx.preamble = preamb_length;
    ptx.invert_pol = invert_enable;
    ptx.size = pl_size;

    if( padvalue > 0 )
    {
        for( i = 0; i < pl_size; ++i )
        {
            ptx.payload[i] = (uint8_t) padvalue & 0xFF;
        }
    }
    else
    {
        memcpy(ptx.payload, payload, pl_size);
    }

    /* Display reminder infos */
    printf( "NOTE: Concentrator board configuration:\n" );

    printf( "  Radio configuration: RX center freq %u Hz, RX bandwidth %u Hz\n", rx_freq_hz, rx_bw_hz );

    if(fskmode)
    {
        ptx.modulation = MOD_FSK;
        ptx.modrate = fskrate * 1e3;
        ptx.f_dev = fskdev;
        printf( "  FSK mode: bitrate %u bps, fdev %u kHz", ptx.modrate, ptx.f_dev);
    }
    else
    {
        ptx.modulation = MOD_LORA;
        ptx.bandwidth = sx1301ar_bw_nb2enum( 1000 * (long)bandwidth_khz );
        ptx.modrate = sx1301ar_sf_nb2enum( spread_factor );
        ptx.coderate = sx1301ar_cr_nb2enum( coding_rate );

        if( loramac_public ) printf( "  LoRaWAN public network mode\n" );
        else printf( "  non-LoRaWAN, or private network mode\n" );

        printf( "  LoRa modulation, BW %ukHz, SF %u, CR %u", bandwidth_khz, spread_factor, coding_rate );
        if( invert_enable ) printf( ", inverted polarity" );
        else printf( ", normal polarity" );
    }

    if( preamb_length == 0 ) printf( ", default preamble length\n" );
    else printf( ", %u symbols preamble\n", preamb_length );

    printf( "  %+idBm TX power, %u bytes payload\n", rf_power, pl_size );

    /* Wait for AGC firmware start (and temp read) */
    sx1301ar_wait_ms( 200 );

    /* Main loop */
    cnt = 0;
    while( (quit_sig != 1) && (exit_sig != 1) )
    {
        if( (quit_sig != 1) && (exit_sig != 1) && ((pkt_repeat == -1) || ((int)cnt < pkt_repeat)) )
        {
            /* Calculate TX frequency */
            if( channels != 1 )
                ptx.freq_hz = start_freq + ((cnt % channels) * chan_offset);

            /* When LBT is enabled, immediate send is not allowed, so we need to set a timestamp to the packet */
            if( lbt_enable == true )
            {
                /* Get the current SX1301 time */
                sx1301ar_get_instcnt( 0, &sx1301_count_us );

                /* Set packet timestamp to current time + few milliseconds */
                ptx.count_us = sx1301_count_us + 50E3;
            }

            /* Send packet */
            x = sx1301ar_send( 0, &ptx );
            if( x != 0 )
            {
                printf( "ERROR: sx1301ar_send failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
                nb_error += 1;
                printf( "NOTE: PER=%.1f%% (sent:%u error:%u)\n", ((float)nb_error/(float)nb_sent)*100, nb_sent, nb_error );
            }
            else
            {
                printf( "X" );
                if( (cnt % 60) == 59 )
                {
                    printf( "\n" );
                }
                else
                {
                    fflush( stdout );
                }
            }

            tx_timeout_us = 0;
            tx_timeout = false;
            sx1301ar_tx_status( 0, &stat );
            while( (stat != TX_FREE) && (quit_sig != 1) && (exit_sig != 1) )
            {
                sx1301ar_wait_us( 100 );
                tx_timeout_us += 100;
                if( (dynatten_timing != 0)
                        && (0 == (tx_timeout_us % (dynatten_timing * 1000))) )
                {
                    x = ad9361_set_atten_offset(0, rf_chain, dynatten_offset);
                    if( x < 0 ) fprintf( stderr, "ERROR: Failed to set AD9361 atten\n");
                }

                if( tx_timeout_us >= 1E6 )
                {
                    printf( "ERROR: TX timeout, aborting...\n" );
                    tx_timeout = true;
                    break;
                }
                sx1301ar_tx_status( 0, &stat );
            }

            if( tx_timeout == true )
            {
                nb_error += 1;
                printf( "NOTE: PER=%.1f%% (sent:%u error:%u)\n", ((float)nb_error/(float)nb_sent)*100, nb_sent, nb_error );
            }
            else
            {
                nb_sent += 1;
                /* If not padvalue, increment Fcnt */
                if(padvalue != 0) ptx.payload[6] = ptx.payload[6] + 1;
            }

            /* Count 1 more TX cycle */
            cnt++;
        }

        fd_stdin = fileno( stdin );

        FD_ZERO( &readfds );
        FD_SET( fileno(stdin), &readfds );

        tv.tv_sec = pkt_delay_ms / 1000;
        tv.tv_usec = (pkt_delay_ms % 1000) * 1000;

        x = select( fd_stdin + 1, &readfds, NULL, NULL, &tv );

         if( x > 0 )
         {
            x = read( fd_stdin, userchar, sizeof(userchar) );
            if( x < 0 ) {
                printf( "Can't read from stdin: %s (%d)\n", strerror( x ), x );
            }
            else if (x > 0) {
                switch( userchar[0] ){
                /* Set Atten value */
                case 'n':
                    txgain_lut.lut[0].ad9361_gain.atten = (uint16_t) strtoul( &userchar[1], NULL, 10 );
                    printf( "Set Atten %u\n", txgain_lut.lut[0].ad9361_gain.atten );

                    sx1301ar_conf_tx_gain( 0, rf_chain, &txgain_lut );
                    cnt=0;
                    break;
                case 'q':
                    quit_sig = 1;
                    break;
                }
            }
         }
    }
    printf( "\n" );

    if( (exit_sig == 1) || ((int)cnt >= pkt_repeat) )
    {
        printf( "NOTE: PER=%.1f%% (sent:%u error:%u)\n", ((float)nb_error/(float)nb_sent)*100, nb_sent, nb_error );

        /* Stop the concentrator */
        x = sx1301ar_stop( 1 );
        if( x != 0 )
        {
            printf( "ERROR: sx1301ar_stop failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
            return EXIT_FAILURE;
        }
        printf( "INFO: concentrator stopped\n" );

        /* Close SPI link */
        x = spi_linuxdev_close( spi1_context );
        if( x != 0 )
        {
            printf( "ERROR: closing SPI failed, returned %i\n", x );
            return EXIT_FAILURE;
        }
        printf( "INFO: SPI link closed\n" );
    }

    printf( "+++  Exiting packet sender +++\n" );
    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void sig_handler( int sigio )
{
    if( sigio == SIGQUIT )
    {
        quit_sig = 1;
    }
    else if( (sigio == SIGINT) || (sigio == SIGTERM) )
    {
        exit_sig = 1;
    }
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static void usage( void )
{
    printf( "~~~ Library version string~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " %s\n", sx1301ar_version_info( 0, NULL, NULL ) );
    printf( "~~~ Interface options ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " -h  print this help\n" );
    printf( " -d <path>  use Linux SPI device driver\n" );
    printf( "  => default path: " DEFAULT_LINUXDEV_PATH "\n" );
    printf( "~~~ Radio and packet options ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " -f <float>        target frequency in MHz\n" );
    printf( " -f <>:<unsigned>  previous + number of channels to jump\n" );
    printf( " -f <>:<>:<float>  previous + explicit offset in MHz between channels\n" );
    printf( " -a <uint>  RF chain [0, 1]\n" );
    printf( " -b <uint>  LoRa bandwidth in kHz [125, 250, 500]\n" );
    printf( " -s <uint>  LoRa Spreading Factor [7-12]\n" );
    printf( " -c <uint>  LoRa Coding Rate [1-4]\n" );
    printf( " -p <int>   RF power (dBm)\n" );
    printf( " -r <uint>  LoRa preamble length (symbols, 0 for default)\n" );
    printf( " -z <uint>  payload size (bytes, <256)\n" );
    printf( " -t <uint>  pause between packets (ms)\n" );
    printf( " -x <int>   nb of times the sequence is repeated (-1 loop until stopped)\n" );
    printf( " -i         send packet using inverted modulation polarity\n" );
    printf( " -m         configure the board in LoRa MAC *public network* mode\n" );
    printf( " -j JFILE   Get TX Gain Table from JSON file JFILE\n" );
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
    printf( " --rfreq           <float>    Radio RX central frequency in MHz\n" );
    printf( " --rbw             <float>    Radio RX bandwidth in MHz\n" );
    printf( " --lbt-freq        <float>    LBT first channel frequency in MHz\n" );
    printf( " --lbt-rssi        <int>      LBT RSSI target in dBm [%d..%d]\n", LBT_RSSI_TARGET_MIN, LBT_RSSI_TARGET_MAX );
    printf( " --lbt-rssi-shift  <uint>     Value to be applied to convert from RSSI in dBm\n" );
    printf( "                               to RSSI register value [0..255] (default=%d)\n", LBT_RSSI_SHIFT );
    printf( " --lbt-sctm        <uint>     LBT scan time for all channels in microseconds [%d, %d]\n", LBT_SCAN_TIME1, LBT_SCAN_TIME2 );
    printf( " --lbt-trtm        <uint>     LBT transmit time for all channels in milliseconds\n" );
    printf( " --lbt-nbch        <uint>     LBT number of channels [1..%d]\n", LBT_NUMBER_OF_CHANNELS_MAX );
    printf( " --atten <uint>  AD9361 attenuation trim, [0:%u]\n", AD9361_ATTEN_MAX );
    printf( " --dac <uint>  AD9361 Aux DAC word, [0:%u]\n", AD9361_AUXDAC_MAX );
    printf( " --vref <uint>  AD9361 Aux DAC Vref LSB, [0:%u]\n", AD9361_VREF_MAX );
    printf( " --pad             <uint>     Padding all payload with <uint> (ex: 0x12, 255, ...)\n" );
    printf( " --payload         <PAYLOAD>  Use PAYLOAD, autodetected size (in hexascci format, ex: 4027262524000000000000)\n" );
    printf( " --dyn-atten       <uint>     Dynamic Atten timing in milliseconds (atten +12 (+3dBm) every <uint> ms)\n" );
    printf( " --dyn-atten       <uint>:<int>     Dynamic Atten timing in milliseconds (atten <int_offset> every <uint> ms)\n" );
    printf( "~~~~ FSK-related options ~~~~\n");
    printf( " --fsk                        Send a FSK packet instead LORA one\n" );
    printf( " --fskdev          <uint>     FSK frequency deviation in kHz [1:250], default:%u (will force --fsk mode)\n", DEFAULT_FDEV_KHZ);
    printf( " --fskrate         <float>    FSK bitrate in kbps [0.5:250], default:%d (will force --fsk mode)\n", DEFAULT_BR_KBPS);
    printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_set_atten_offset( uint8_t brd, uint8_t rf_chain, int16_t atten_offset)
{
    uint8_t u8val1, u8val2;
    uint16_t atten = 0;

    /* Check input parameters */
    if( brd >= SX1301AR_MAX_BOARD_NB )
    {
        fprintf( stderr, "ERROR: wrong board number (%u)\n", brd );
        return -1;
    }
    if( (rf_chain != 0) && (rf_chain != 1) )
    {
        fprintf( stderr, "ERROR: wrong rf_chain (%u)\n", rf_chain );
        return -1;
    }

    /* Get SPI control of AD9361 (instead SX1301) */
    sx1301ar_reg_brd_w( 0, BREG_AD9361_SPI_HOST_CTRL, 1 );

    /* Read attenuation register (9-bits) */
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, &u8val1 );
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, &u8val2 );
    atten = (u8val1 + (u8val2 << 8));

    /* Apply offset */
    atten = atten + atten_offset;

    /* Write attenuation register (9-bits) */
    sx1301ar_ad9361_w( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, (uint8_t)(atten & 0xFF) );
    sx1301ar_ad9361_w( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, (uint8_t)((atten >> 8) & 1) );

    /* Read back attenuation register (9-bits) */
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, &u8val1 );
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, &u8val2 );

    /* Give back SPI control of AD9361 to SX1301 */
    sx1301ar_reg_brd_w( 0, BREG_AD9361_SPI_HOST_CTRL, 0 );

    /* Check attenuation value */
    if( (u8val1 + (u8val2 << 8)) != atten )
    {
        fprintf( stderr, "ERROR: failed to set attenuation %d (read:%u)\n", atten, u8val1 + (u8val2 << 8) );
        return -1;
    }
    else
    {
        printf( "INFO: Configuring AD9361 with atten:%d (%.2f dB)\n", atten, atten * 0.25);
    }

    return 0;
}

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

int extract_rfchain_configuration( JSON_Array * conf_rfchain_array, sx1301ar_board_cfg_t *cfg_brd )
{
    int i, j;
    JSON_Object * conf_txgain_obj;
    JSON_Object * conf_rfchain_obj;
    JSON_Array * conf_txlut_array;
    JSON_Value * val;

    if( conf_rfchain_array != NULL )
    {
        for( i = 0; i < (int) json_array_get_count( conf_rfchain_array ); i++ )
        {
            /* Sanity check */
            if( i >= SX1301AR_BOARD_RFCHAIN_NB )
            {
                printf( "ERROR: rf_chain %d not supported, skip it\n", i );
                return EXIT_FAILURE;
            }

            /* Get RF chain conf object */
            conf_rfchain_obj = json_array_get_object( conf_rfchain_array, i );

            /* Parse configuration */
            cfg_brd->rf_chain[i].rx_enable = (bool) json_object_get_boolean( conf_rfchain_obj, "rx_enable" );
            cfg_brd->rf_chain[i].tx_enable = (bool) json_object_get_boolean( conf_rfchain_obj, "tx_enable" );

            /* set configuration for tx gains */
            conf_txlut_array = json_object_get_array( conf_rfchain_obj, "tx_lut" );
            if( conf_txlut_array != NULL )
            {
                cfg_brd->rf_chain[i].tx_lut.size = json_array_get_count( conf_txlut_array );
                printf( "INFO: rf_chain %d: Configuring Tx LUT with %u indexes\n", i, cfg_brd->rf_chain[i].tx_lut.size );
                for( j = 0; j < (int) cfg_brd->rf_chain[i].tx_lut.size; j++ )
                {
                    /* Sanity check */
                    if( j >= SX1301AR_BOARD_MAX_LUT_NB )
                    {
                        printf( "ERROR: TX LUT index %d not supported, skip it\n", j );
                        break;
                    }

                    /* Get TX gain object from LUT */
                    conf_txgain_obj = json_array_get_object( conf_txlut_array, j );

                    /* rf power */
                    val = json_object_dotget_value( conf_txgain_obj, "rf_power" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].rf_power = (int8_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "rf_power", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].rf_power = 0;
                    }

                    /* FPGA dig gain */
                    val = json_object_dotget_value( conf_txgain_obj, "fpga_dig_gain" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].fpga_dig_gain = (uint8_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "fpga_dig_gain", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].fpga_dig_gain = 0;
                    }

                    /* AD9361 atten */
                    val = json_object_get_value( conf_txgain_obj, "ad9361_atten" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.atten = (uint16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_atten", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.atten = 0;
                    }

                    /* AD9361 auxdac vref */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_auxdac_vref" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_vref = (uint8_t) json_value_get_number( val );
                    }
                    else
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_vref = 0;
                    }

                    /* AD9361 auxdac word */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_auxdac_word" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_word = (uint16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_auxdac_word", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_word = 0;
                    }

                    /* AD9361 temperature compensation coeff A */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_tcomp_coeff_a" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_a = (int16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_tcomp_coeff_a", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_a = 0;
                    }

                    /* AD9361 temperature compensation coeff B */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_tcomp_coeff_b" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_b = (int16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_tcomp_coeff_b", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_b = 0;
                    }
                }
            }
        }
    }
    else
    {
        printf( "WARNING: no rf_chain configured\n" );
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

/* --- EOF ------------------------------------------------------------------ */
