/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    LoRa Gateway v2 spectral scan utility

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Matthieu Leurent
*/


/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <string.h>     /* strcmp */
#include <getopt.h>     /* getopt_long */

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"
#include "sx1301ar_radio.h"

#include "spi_linuxdev.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define DEFAULT_RADIO_RX_FREQ   866500000   /* radio RX central frequency, Hz */
#define DEFAULT_RADIO_RX_BANDWIDTH 4E6
#define DEFAULT_START_FREQ      863000000   /* start frequency, Hz */
#define DEFAULT_STOP_FREQ       870000000   /* stop frequency, Hz */
#define DEFAULT_STEP_FREQ       50000       /* frequency step, Hz */
#define DEFAULT_RSSI_PTS        8192        /* number of RSSI reads */
#define DEFAULT_RSSI_RATE_DIV   9           /* RSSI sampling rate = 125kHz/(div+1) */
#define DEFAULT_RSSI_OFFSET     -143        /* RSSI offset of AD9361 -> FPGA rf chain (no SX1301) */
#define DEFAULT_LOG_NAME        "rssi_histogram"

#define LINUXDEV_PATH_DEFAULT   "/dev/spidev0.0"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );


/* -------------------------------------------------------------------------- */
/* --- USEFULL FUNCTIONS --------------------------------------------- */

static int start_board(uint32_t rx_freq_hz, uint32_t rx_bw, uint8_t rfchain)
{
	sx1301ar_board_cfg_t cfg = sx1301ar_init_board_cfg();
	int ret;

	/* Dummy parameters, just need to set SPI callbacks */
	cfg.spi_read = &spi1_read;
	cfg.spi_write = &spi1_write;
	cfg.board_type = BRD_MASTER;
	cfg.nb_chip = 2;
	cfg.rx_freq_hz = rx_freq_hz;
	cfg.rx_bw_hz = rx_bw;
	cfg.rf_chain[rfchain].rx_enable = true;

	if (sx1301ar_conf_board(0, &cfg) < 0) {
		printf("Failed to configure board: %s\n",
		      sx1301ar_err_message(sx1301ar_errno));
		return -1;

	}

    /* Start the concentrator */
    ret = sx1301ar_start( 1 );
    if( ret != 0 )
    {
        printf( "ERROR: sx1301ar_start failed ; %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( 1 );
        return -2;
    }


	return 0;
}


/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
    int i, j, k; /* loop and temporary variables */
    int x; /* return code for SX1301ar driver functions */
    int32_t val;

    /* SPI interfaces */
    const char spi1_path_default[] = LINUXDEV_PATH_DEFAULT;
    const char * spi1_path = spi1_path_default;

    /* Parameter parsing */
    int option_index = 0;
    static struct option long_options[] = {
        { "rfreq", 1, 0, 0 },
        { "rxbw", 1, 0, 0 },
        { "rfchain", 1, 0, 0 },
        { 0, 0, 0, 0 }
    };
    double arg_lf[3] = { 0, 0, 0 };
    unsigned arg_u = 0;
    int arg_i = 0;
    char arg_s[64];
    int valid_retry;

    /* Application parameters */
    uint32_t rx_freq_hz = DEFAULT_RADIO_RX_FREQ;
    uint32_t rx_bw_hz = DEFAULT_RADIO_RX_BANDWIDTH;
    uint32_t start_freq = DEFAULT_START_FREQ;
    uint32_t stop_freq = DEFAULT_STOP_FREQ;
    uint32_t step_freq = DEFAULT_STEP_FREQ;
    uint16_t rssi_pts = DEFAULT_RSSI_PTS;
    uint16_t rssi_rate_div = DEFAULT_RSSI_RATE_DIV;
    int16_t rssi_offset = DEFAULT_RSSI_OFFSET;
    char log_file_name[64] = DEFAULT_LOG_NAME;
    FILE * log_file = NULL;

    int32_t histo_valid = 0;
    int freq_nb;
    uint32_t freq;
    int32_t if_freq;
    uint8_t read_burst[512];
    uint16_t rssi_histo;
    uint16_t rssi_cumu;
    float rssi_thresh[] = { 0.1, 0.3, 0.5, 0.8, 1 };

    uint8_t rfchain = 0;
    uint8_t strip_zero = 0;
    uint8_t standalone = 0;

    /* Parse command line options */
    while( (i = getopt_long( argc, argv, "hd:f:n:r:l:o:s0", long_options, &option_index )) != -1 )
    {
        switch( i )
        {
        case 'h':
            printf( "~~~ Library version string~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
            printf( " %s\n", sx1301ar_version_info( 0, NULL, NULL ) );
            printf( "~~~ Interface options ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
            printf( " -h            print this help\n" );
            printf( " -d <path>     use Linux SPI device driver\n" );
            printf( "               => default path: " LINUXDEV_PATH_DEFAULT "\n" );
            printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
            printf( " -f <float>:<float>:<float>  Frequency vector to scan in MHz, start:step:stop, default:%.1lf:%.2lf:%.1lf\n", DEFAULT_START_FREQ/1E6, DEFAULT_STEP_FREQ/1E6, DEFAULT_STOP_FREQ/1E6 );
            printf( " -n <uint>     Total number of RSSI points, [1,65535], default: 8192\n" );
            printf( " -r <uint>     Divide factor of RSSI sampling rate, 125kHz/(div+1), [0,65535], default:9\n" );
            printf( " -o <int>      RSSI offset in dBm, default:-143\n" );
            printf( " -l <char>     Log file name (default: "DEFAULT_LOG_NAME"\n" );
            printf("  -s            Standalone mode (Start LNA - no need of pkt_logger)\n");
            printf("  -0            Don't save zero values\n");
            printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
            printf( " --rxbw    <float>    Radio RX bandwidth in MHz, default:%.1lf\n", DEFAULT_RADIO_RX_BANDWIDTH/1E6 );
            printf( " --rfreq   <float>    Radio RX central frequency in MHz, default:%.1lf\n", DEFAULT_RADIO_RX_FREQ/1E6 );
            printf( " --rfchain <uint>     Radio RX chain (0-1), default: 0\n" );
            printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
            return EXIT_SUCCESS;
            break;

        case 'd': /* -d <optional path>  use Linux SPI device driver */
            if( optarg != NULL )
            {
                spi1_path = optarg;
            }
            break;

        case 'f': /* -f <float>:<float>:<float>  Frequency vector to scan in MHz, start:step:stop */
            j = sscanf( optarg, "%lf:%lf:%lf", &arg_lf[0], &arg_lf[1], &arg_lf[2] );
            if( (j!=3) || (arg_lf[0] < 0) )
            {
                printf( "ERROR: argument parsing of -f argument. Use -h to print help\n" );
                return EXIT_FAILURE;
            }
            else
            {
                start_freq = (uint32_t)((arg_lf[0] * 1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
                step_freq = (uint32_t)((arg_lf[1] * 1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
                stop_freq = (uint32_t)((arg_lf[2] * 1e6) + 0.5); /* .5 Hz offset to get rounding instead of truncating */
            }
            break;

        case 'n': /* -n <uint>  Total number of RSSI points, [1,65535] */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u < 1) || (arg_u > 65535) )
            {
                printf( "ERROR: argument parsing of -n argument. Use -h to print help\n" );
                return EXIT_FAILURE;
            }
            else
            {
                rssi_pts = (uint16_t)arg_u;
            }
            break;

        case 'r': /* -r <uint>  Divide factor of RSSI sampling rate, 1MHz/(div+1), [0,65535] */
            j = sscanf( optarg, "%u", &arg_u );
            if( (j != 1) || (arg_u > 65535) )
            {
                printf( "ERROR: argument parsing of -r argument. Use -h to print help\n" );
                return EXIT_FAILURE;
            }
            else
            {
                rssi_rate_div = (uint16_t)arg_u;
            }
            break;

        case 'l': /* -l <char>  Log file name */
            j = sscanf( optarg, "%63s", arg_s );
            if( j != 1 )
            {
                printf( "ERROR: argument parsing of -l argument. Use -h to print help\n" );
                return EXIT_FAILURE;
            }
            else
            {
                sprintf(log_file_name, "%s", arg_s);
            }
            break;

        case 'o': /* -o <int>  RSSI offset */
            j = sscanf( optarg, "%i", &arg_i );
            if( (j != 1) )
            {
                printf( "ERROR: argument parsing of -o argument. Use -h to print help\n" );
                return EXIT_FAILURE;
            }
            else
            {
                rssi_offset = (int16_t)arg_i;
            }
            break;

        case 's':
            standalone = 1;
            break;

        case '0':
            strip_zero = 1;
            break;

        case 0:
            if( strcmp( long_options[option_index].name, "rfreq" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_lf[0] );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --rfreq argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {

                    rx_freq_hz = (uint32_t)( arg_lf[0] * 1e6 );
                    if(standalone == 0)
                    {
                        printf( " WARNING : Make sure that board_rx_freq=%u in pkt_logger config\n",
                                rx_freq_hz);
                    }
                }
            }
            else if( strcmp( long_options[option_index].name, "rxbw" ) == 0 )
            {
                i = sscanf( optarg, "%lf", &arg_lf[0] );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --rxbw argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    rx_bw_hz = (uint32_t)( arg_lf[0] * 1e6 );
                }
            }
            else if( strcmp( long_options[option_index].name, "rfchain" ) == 0 )
            {
                i = sscanf( optarg, "%u", &arg_u );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of --rxbw argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    rfchain = arg_u;
                }
            }
            break;

        default:
            printf( "ERROR: argument parsing options. Use -h to print help\n" );
            return EXIT_FAILURE;
        }
    }

    /* Start message */
    printf( "+++ Start RSSI histogram of 'SX1301 array' channels +++\n" );
    printf( "SX1301 array library version string: %s\n", sx1301ar_version_info( 0, NULL, NULL ) );

    /* Open SPI link */
    i = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( i != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i\n", i );
        return EXIT_FAILURE;
    }

    /* Register SPI callbacks */
    x = sx1301ar_reg_setup( 0, BRD_MASTER, spi1_read, spi1_write );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_reg_setup failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    /* Check FPGA version register */
    x = sx1301ar_reg_brd_r( 0, BREG_VERSION, &val );
    SX1301AR_ERR_PROPAGATE( x );
    if( val == 0 )
    {
        fprintf( stderr, "[BRD%d] FPGA version = %i\n", 0, val ); // DEBUG
        SX1301AR_ERR_RETURN( ERR_FPGA_VZERO );
    }
    else if( val != SX1301AR_BOARD_VERSION_MASTER )
    {
        fprintf( stderr, "[BRD%d] FPGA version = %i\n", 0, val ); // DEBUG
        SX1301AR_ERR_RETURN( ERR_FPGA_VERSION );
    }
    printf( "FPGA version: %d\n", val );

    if (standalone) {
        printf("Starting LNA\n");
        x = start_board(rx_freq_hz, rx_bw_hz, rfchain);
        if(x)
            return EXIT_FAILURE;
    }

    /* Configure RSSI histogram */
    sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_RATE_DIV, (int32_t)rssi_rate_div );
    sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_LEN, (int32_t)rssi_pts );
    sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_SELECT_RF_PATH, (int32_t)rfchain);

    /* create log file */
    strcat( log_file_name,".csv" );
    log_file = fopen( log_file_name, "w" );
    if( log_file == NULL )
    {
        printf( "ERROR: impossible to create log file %s\n", log_file_name );
        return EXIT_FAILURE;
    }

    /* Print parameters */
    printf( "File        : %s\n", log_file_name );
    printf( "Freq start  : %d Hz\n", start_freq );
    printf( "Freq stop   : %d Hz\n", stop_freq );
    printf( "Freq step   : %d Hz\n", step_freq );
    sx1301ar_reg_brd_r( 0, BREG_RSSI_HISTO_LEN, &val );
    printf( "Pts per step: %d\n", val );
    sx1301ar_reg_brd_r( 0, BREG_RSSI_HISTO_RATE_DIV, &val );
    printf( "Rate        : %1.0f Hz\n", 125e3/((float)val+1) );
    printf( "RSSI offset : %d dBm\n", rssi_offset );
    printf( "Radio RX center frequency: %u Hz\n", rx_freq_hz );

    /* Number of frequency steps */
    freq_nb = (int)( (stop_freq - start_freq) / step_freq ) + 1;

    /* Main loop */
    j = 0;
    while( j < freq_nb )
    {
        /* Current frequency */
        freq = start_freq + j * step_freq;
        if_freq = (int32_t)( -( (float)freq - (float)rx_freq_hz ) / 1e6 * 512 - 0.5 );
        /* The IF register is 13-bits, so if the calculated if_freq exceed the range, we exit */
        if( (if_freq < -4095) || (if_freq > 4095) )
        {
            printf( "ERROR: IF frequency is out of range (%d)\n", if_freq );
            break;
        }
        sx1301ar_reg_brd_w( 0, BREG_IF_FREQ_RSSI, if_freq );

        /* Start histogram */
        sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_START, 1 );

        /* Wait until rssi_pts have been processed */
        valid_retry = 0;
        do
        {
            sx1301ar_wait_us( 100000 );
            x = sx1301ar_reg_brd_r( 0, BREG_RSSI_HISTO_VALID, &histo_valid );
            if( x != 0 )
            {
                printf( "ERROR: failed to read histogram value\n" );
            }
            valid_retry += 1;
        }
        while( (histo_valid != 1) && (valid_retry < 100) );

        /* Unset start bit */
        sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_START, 0 );

        /* Check if we could get a valid RSSI histogram */
        if( histo_valid == 0 )
        {
            printf( "TIMEOUT: failed to get valid RSSI histogram for %d, retry...\n", freq );
            continue;
        }

        /* Read back histogram */
        sx1301ar_reg_brd_w( 0, BREG_RSSI_HISTO_RAM_ADDR, 0 );
        sx1301ar_reg_brd_rb( 0, BREG_RSSI_HISTO_RAM_DATA, read_burst, 512 );

        printf( "%d", freq );
        fprintf( log_file, "%d", freq );
        rssi_cumu = 0;
        k = 0;
        for( i = 0; i < 256; i++ )
        {
            rssi_histo = read_burst[2*i] | (read_burst[2*i+1] << 8);
            if (strip_zero == 0 || rssi_histo != 0)
            {
               fprintf( log_file, ",%d,%d", i+rssi_offset, rssi_histo );
            }
            rssi_cumu += rssi_histo;
            if( rssi_cumu > rssi_thresh[k]*rssi_pts )
            {
                printf( "  %d%%<%4d", (uint16_t)(rssi_thresh[k]*100), i+rssi_offset );
                k++;
            }
        }
        fprintf( log_file, "\n" );
        printf( "\n" );

        /* increment while loop counter */
        j += 1;
    }
    fclose( log_file );

    if (standalone)
        sx1301ar_stop(1);

    /* Close SPI link */
    i = spi_linuxdev_close( spi1_context );
    if( i != 0 )
    {
        printf( "ERROR: closing SPI failed, returned %i\n", i );
        return EXIT_FAILURE;
    }
    printf( "INFO: SPI link closed\n" );

    printf( "+++  Exiting RSSI histogram program +++\n" );
    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */
