/*
  ______                              _
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2014 Semtech-Cycleo

Description:
    LoRa Gateway v2 tx continuous utility

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Matthieu Leurent
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

/* Fix an issue between POSIX and C99 */
#if __STDC_VERSION__ >= 199901L
    #define _XOPEN_SOURCE 600
#else
    #define _XOPEN_SOURCE 500
#endif

#include <stdint.h>     /* C99 types */
#include <stdio.h>      /* NULL printf */
#include <stdlib.h>     /* EXIT atoi */
#include <unistd.h>     /* getopt */
#include <signal.h>     /* sigaction */
#include <string.h>     /* strcmp */
#include <getopt.h>     /* getopt_long */
#include <math.h>       /* sin, cos, sqrt */

#include "sx1301ar_aux.h"
#include "sx1301ar_err.h"
#include "sx1301ar_hal.h"
#include "sx1301ar_reg.h"
#include "sx1301ar_radio.h"

#include "spi_linuxdev.h"

#include "parson.h"

/* -------------------------------------------------------------------------- */
/* --- MACROS & CONSTANTS --------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))

#define DEFAULT_RADIO_RX_FREQ   866500000
#define DEFAULT_RADIO_RX_BW     7000000
#define DEFAULT_FREQ_HZ         869525000
#define DEFAULT_RF_CHAIN        0
#define DEFAULT_DIG_GAIN        5
#define DEFAULT_AD9361_ATTEN    87
#define DEFAULT_AUXDAC_VREF     3
#define DEFAULT_AUXDAC_WORD     770
#define DEFAULT_RF_POWER        14
#define DEFAULT_MODULATION      "LORA"
#define DEFAULT_SF              7
#define DEFAULT_BW_KHZ          125
#define DEFAULT_BR_KBPS         50
#define DEFAULT_FDEV_KHZ        25
#define DEFAULT_BT              2
#define DEFAULT_LINUXDEV_PATH   "/dev/spidev32766.0"

#define JSON_SX1301AR_CONF_OBJ  "SX1301_array_conf"

/* -------------------------------------------------------------------------- */
/* --- GLOBAL VARIABLES ----------------------------------------------------- */

/* Signal handling variables */
static int exit_sig = 0; /* 1 -> application terminates cleanly (shut down hardware, close open files, etc) */
static int quit_sig = 0; /* 1 -> application terminates without shutting down the hardware */

/* SPI interfaces */
int spi1_context = -1; /* context for linuxdev SPI driver (file descriptor) */

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DECLARATION --------------------------------------------- */

static void sig_handler( int sigio );

/* Wrappers around linuxdev SPI functions with explicit (global variables) SPI targets */
static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status );
static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status );

static int ad9361_atten_set_val( uint8_t brd, uint8_t rf_chain, uint16_t atten, int16_t tcomp_coeff_a, int16_t tcomp_coeff_b, uint8_t ad9361_temp_ref );
static int ad9361_read_temperature( uint8_t brd, uint8_t * temperature_code );

int extract_rfchain_configuration( JSON_Array * conf_rfchain_array, sx1301ar_board_cfg_t *cfg_brd );

/* -------------------------------------------------------------------------- */
/* --- MAIN FUNCTION -------------------------------------------------------- */

int main( int argc, char ** argv )
{
    static struct sigaction sigact; /* SIGQUIT&SIGINT&SIGTERM signal handling */

    int i; /* loop and temporary variables */
    int x; /* return code for SX1301ar driver functions */
    int32_t val;

    /* SPI interfaces */
    const char spi1_path_default[] = DEFAULT_LINUXDEV_PATH;
    const char * spi1_path = spi1_path_default;

    /* Configuration structures */
    sx1301ar_board_cfg_t cfg_brd = sx1301ar_init_board_cfg( );
    sx1301ar_tx_gain_t g = sx1301ar_init_tx_gain( );

    /* Parameter parsing */
    int option_index = 0;
    static struct option long_options[] = {
        {"rfreq", 1, 0, 0},
        {"rbw", 1, 0, 0},
        {"rf", 1, 0, 0},
        {"dig", 1, 0, 0},
        {"atten", 1, 0, 0},
        {"vref", 1, 0, 0},
        {"dac", 1, 0, 0},
        {"tmp", 1, 0, 0},
        {"tca", 1, 0, 0},
        {"tcb", 1, 0, 0},
        {"mod", 1, 0, 0},
        {"sf", 1, 0, 0},
        {"bw", 1, 0, 0},
        {"br", 1, 0, 0},
        {"fdev", 1, 0, 0},
        {"bt", 1, 0, 0},
        {0, 0, 0, 0}
    };
    unsigned int arg_u;
    int arg_i;
    float arg_f;
    double arg_lf;
    char arg_s[64];
    char userchar[10];
    char *cret = NULL;
    uint8_t u8val;
    uint8_t u8val2;

    /* Application parameters */
    uint32_t rx_freq_hz = DEFAULT_RADIO_RX_FREQ;
    uint32_t rx_bw_hz = DEFAULT_RADIO_RX_BW;
    uint32_t f_hz = DEFAULT_FREQ_HZ;
    uint8_t rf_chain = DEFAULT_RF_CHAIN;
    char mod[64] = DEFAULT_MODULATION;
    uint8_t sf = DEFAULT_SF;
    unsigned int bw_khz = DEFAULT_BW_KHZ;
    float br_kbps = DEFAULT_BR_KBPS;
    uint8_t fdev_khz = DEFAULT_FDEV_KHZ;
    uint8_t bt = DEFAULT_BT;
    g.fpga_dig_gain = DEFAULT_DIG_GAIN;
    g.ad9361_gain.atten = DEFAULT_AD9361_ATTEN;
    g.ad9361_gain.auxdac_vref = DEFAULT_AUXDAC_VREF;
    g.ad9361_gain.auxdac_word = DEFAULT_AUXDAC_WORD;
    g.ad9361_tcomp.coeff_a = 0;
    g.ad9361_tcomp.coeff_b = 0;

    sx1301ar_tx_pkt_t p;
    sx1301ar_tstat_t tx_status;

    JSON_Value * root_val;
    JSON_Array * conf_sx1301_array = NULL;
    JSON_Array * conf_rfchain_array = NULL;
    JSON_Object * root = NULL;
    JSON_Object * brd_conf_obj = NULL;
    int8_t rf_power = DEFAULT_RF_POWER;

    /* Parse command line options */
    while( (i = getopt_long( argc, argv, "hf:d:j:p:", long_options, &option_index) ) != -1 )
    {
        switch( i )
        {
            case 'h':
                printf( "~~~ Library version string~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
                printf( " %s\n", sx1301ar_version_info( 0, NULL, NULL ) );
                printf( "~~~ Interface options ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
                printf( " -h  print this help\n" );
                printf( " -f <float>    RF frequency in MHz\n" );
                printf( " -d <path>     use Linux SPI device driver. Default path: " DEFAULT_LINUXDEV_PATH "\n" );
                printf( " -j JFILE   Get TX Gain Table from JSON file JFILE\n" );
                printf( " -p <int>   RF power (dBm)\n" );
                printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
                printf( " --rfreq <float>   Radio RX central frequency in MHz\n" );
                printf( " --rbw   <float>   Radio RX bandwidth in MHz\n" );
                printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
                printf( " --rf    <uint>    RF chain, [0:1]\n" );
                printf( " --dig   <uint>    FPGA digital gain trim, [0:13]\n" );
                printf( " --atten <uint>    AD9361 attenuation trim, [0:%d]\n", AD9361_TXGAIN_ATTEN_MAX );
                printf( " --vref  <uint>    AD9361 Aux DAC Vref LSB, [0:%d]\n", AD9361_TXGAIN_VREF_MAX );
                printf( " --dac   <uint>    AD9361 Aux DAC word, [0:%d]\n", AD9361_TXGAIN_AUXDAC_MAX );
                printf( " --tmp   <int>     AD9361 reference temperature code\n" );
                printf( " --tca   <int>     AD9361 temperature compensation coeff A\n" );
                printf( " --tcb   <int>     AD9361 temperature compensation coeff B\n" );
                printf( " --mod   <char>    Modulation type ['LORA','FSK','CW']\n" );
                printf( " --sf    <uint>    LoRa Spreading Factor, [7:12]\n" );
                printf( " --bw    <uint>    LoRa bandwidth in kHz, [125,250,500]\n" );
                printf( " --br    <float>   FSK bitrate in kbps, [0.5:250]\n" );
                printf( " --fdev  <uint>    FSK frequency deviation in kHz, [1:250]\n" );
                printf( " --bt    <uint>    FSK gaussian filter BT trim, [0:3]\n" );
                printf( "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n" );
                return EXIT_SUCCESS;
                break;

            case 'f':
                i = sscanf( optarg, "%lf", &arg_lf );
                if( i != 1 )
                {
                    printf( "ERROR: argument parsing of -f argument. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    f_hz = (uint32_t)( arg_lf * 1e6 );
                }
                break;

            case 'd': /* -d <path>  use Linux SPI device driver */
                if( optarg != NULL )
                {
                    spi1_path = optarg;
                }
                break;

            case 'j': /* -j  Get Calibration (Lut table) from a JSON file */
                printf( "INFO: found calibration file %s\n", optarg );
                root_val = json_parse_file_with_comments( optarg );
                root = json_value_get_object( root_val );
                if( root == NULL )
                {
                    printf( "ERROR: %s is not a valid JSON file\n", optarg );
                    return EXIT_FAILURE;
                }
                conf_sx1301_array = json_object_get_array( root, JSON_SX1301AR_CONF_OBJ );
                if( conf_sx1301_array != NULL )
                {
                    brd_conf_obj = json_array_get_object( conf_sx1301_array, 0 );
                    if( brd_conf_obj == NULL )
                    {
                        printf( "ERROR: %s array does not include item no%u\n", JSON_SX1301AR_CONF_OBJ, 0 );
                        return EXIT_FAILURE;
                    }
                }
                else
                {
                    printf( "ERROR: configuration file does not contain a JSON object named %s\n", JSON_SX1301AR_CONF_OBJ );
                    return EXIT_FAILURE;
                }
                break;

            case 'p': /* -p <int>  RF power (dBm) */
                i = sscanf( optarg, "%i", &arg_i );
                if( (i != 1) || (arg_i < -60) || (arg_i > 60) )
                {
                    printf( "ERROR: argument parsing of -p argument\n" );
                    return EXIT_FAILURE;
                }
                else
                {
                    rf_power = (int8_t)arg_i;
                }
                break;

            case 0:
                if( strcmp( long_options[option_index].name, "rf" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > 1) )
                    {
                        printf( "ERROR: argument parsing of --rf argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        rf_chain = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "rfreq" ) == 0 )
                {
                    i = sscanf( optarg, "%lf", &arg_lf );
                    if( i != 1 )
                    {
                        printf( "ERROR: argument parsing of --rfreq argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        rx_freq_hz = (uint32_t)( arg_lf * 1e6 );
                    }
                }
                else if( strcmp( long_options[option_index].name, "rbw" ) == 0 )
                {
                    i = sscanf( optarg, "%lf", &arg_lf );
                    if( i != 1 )
                    {
                        printf( "ERROR: argument parsing of --rbw argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        rx_bw_hz = (uint32_t)( arg_lf * 1e6 );
                    }
                }
                else if( strcmp( long_options[option_index].name, "dig" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > 13) )
                    {
                        printf( "ERROR: argument parsing of --dig argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.fpga_dig_gain = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "atten" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > AD9361_TXGAIN_ATTEN_MAX) )
                    {
                        printf( "ERROR: argument parsing of --atten argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.ad9361_gain.atten = (uint16_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "vref" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > AD9361_TXGAIN_VREF_MAX) )
                    {
                        printf( "ERROR: argument parsing of --vref argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.ad9361_gain.auxdac_vref = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "dac" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > AD9361_TXGAIN_AUXDAC_MAX) )
                    {
                        printf( "ERROR: argument parsing of --dac argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.ad9361_gain.auxdac_word = (uint16_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "tmp" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) )
                    {
                        printf( "ERROR: argument parsing of --tmp argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        cfg_brd.ad9361_temp_ref = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "tca" ) == 0 )
                {
                    i = sscanf( optarg, "%d", &arg_i );
                    if( (i != 1) )
                    {
                        printf( "ERROR: argument parsing of --tca argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.ad9361_tcomp.coeff_a = (int16_t)arg_i;
                    }
                }
                else if( strcmp( long_options[option_index].name, "tcb" ) == 0 )
                {
                    i = sscanf( optarg, "%d", &arg_i );
                    if( (i != 1) )
                    {
                        printf( "ERROR: argument parsing of --tcb argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        g.ad9361_tcomp.coeff_b = (int16_t)arg_i;
                    }
                }
                else if( strcmp( long_options[option_index].name, "mod" ) == 0 )
                {
                    i = sscanf( optarg, "%63s", arg_s );
                    if( (i != 1) || ((strcmp(arg_s,"LORA") != 0) && (strcmp(arg_s,"FSK") != 0)  && (strcmp(arg_s,"CW") != 0)) )
                    {
                        printf( "ERROR: argument parsing of --mod argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        sprintf( mod, "%s", arg_s );
                    }
                }
                else if( strcmp( long_options[option_index].name, "sf" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u < 7) || (arg_u > 12) )
                    {
                        printf( "ERROR: argument parsing of --sf argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        sf = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "bw" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || ((arg_u != 125) && (arg_u != 250) && (arg_u != 500)) )
                    {
                        printf( "ERROR: argument parsing of --bw argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        bw_khz = arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "br" ) == 0 )
                {
                    i = sscanf( optarg, "%f", &arg_f );
                    if( (i != 1) || (arg_f < 0.5) || (arg_f > 250) )
                    {
                        printf( "ERROR: argument parsing of --br argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        br_kbps = arg_f;
                    }
                }
                else if( strcmp( long_options[option_index].name, "fdev" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u < 1) || (arg_u > 250) )
                    {
                        printf( "ERROR: argument parsing of --fdev argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        fdev_khz = (uint8_t)arg_u;
                    }
                }
                else if( strcmp( long_options[option_index].name, "bt" ) == 0 )
                {
                    i = sscanf( optarg, "%u", &arg_u );
                    if( (i != 1) || (arg_u > 3) )
                    {
                        printf( "ERROR: argument parsing of --bt argument. Use -h to print help\n" );
                        return EXIT_FAILURE;
                    }
                    else
                    {
                        bt = (uint8_t)arg_u;
                    }
                }
                else
                {
                    printf( "ERROR: argument parsing options. Use -h to print help\n" );
                    return EXIT_FAILURE;
                }
                break;

        default:
            printf( "ERROR: argument parsing options. Use -h to print help\n" );
            return EXIT_FAILURE;
        }
    }

    /* Recap all settings */
    printf( "SX1301 array library version string: %s\n", sx1301ar_version_info( 0, NULL, NULL ) );
    printf( "RF frequency: %4.3f MHz (%u Hz)\n", (float)f_hz/1e6, f_hz );
    printf( "RF path: %d\n", rf_chain );
    if( strcmp( mod, "LORA" ) == 0 )
    {
        printf( "Modulation: LORA SF:%u BW:%u kHz\n", sf, bw_khz );
    }
    else if( strcmp( mod, "FSK" ) == 0 )
    {
        printf( "Modulation: FSK BR:%3.3f kbps FDEV:%d kHz BT:%d\n", br_kbps, fdev_khz, bt );
    }
    else if( strcmp( mod, "CW" ) == 0 )
    {
        printf( "Modulation: CW\n" );
    }

    /* Configure signal handling */
    sigemptyset( &sigact.sa_mask );
    sigact.sa_flags = 0;
    sigact.sa_handler = sig_handler;
    sigaction( SIGQUIT, &sigact, NULL );
    sigaction( SIGINT, &sigact, NULL );
    sigaction( SIGTERM, &sigact, NULL );

    /* Open SPI link */
    x = spi_linuxdev_open( spi1_path, -1, &spi1_context );
    if( x != 0 )
    {
        printf( "ERROR: opening SPI failed, returned %i\n", x );
        return EXIT_FAILURE;
    }

    /* Configure SPI devices to master/slave modes (Needed for ATMEL CPU only)
         SPI 0: HOST <-> FPGA
         SPI 1: HOST/DSP <-> Flash
    */
    x = spi_set_mode( 0, SPI_MODE_MASTER );
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", x );
        return EXIT_FAILURE;
    }
    x = spi_set_mode( 1, SPI_MODE_SLAVE ); /* So that DSP can access the flash for booting */
    if( x != 0 )
    {
        printf( "ERROR: setting SPI mode failed, returned %i\n", x );
        return EXIT_FAILURE;
    }

    /* Configure the board */
    cfg_brd.spi_read = &spi1_read;
    cfg_brd.spi_write = &spi1_write;
    cfg_brd.board_type = BRD_MASTER;
    cfg_brd.nb_chip = 1; /* enough for TX */
    cfg_brd.rx_freq_hz = rx_freq_hz;
    cfg_brd.rx_bw_hz = rx_bw_hz;
    cfg_brd.full_duplex = false;
    cfg_brd.rf_chain[rf_chain].tx_enable = true;
    /* Set gains by using the first index of gain LUT */
    g.rf_power = rf_power;
    cfg_brd.rf_chain[rf_chain].tx_lut.size = 1;
    cfg_brd.rf_chain[rf_chain].tx_lut.lut[0] = g;

    if(brd_conf_obj != NULL)
    {
        /* RF chains configuration */
        conf_rfchain_array = json_object_get_array( brd_conf_obj, "rf_chain_conf" );
        x = extract_rfchain_configuration( conf_rfchain_array, &cfg_brd );
        /* Search requested power in TX gain LUT */
        for( i = 0; i < cfg_brd.rf_chain[rf_chain].tx_lut.size; i++ )
        {
            if(cfg_brd.rf_chain[rf_chain].tx_lut.lut[i].rf_power > rf_power)
            {
                break;
            }
            g = cfg_brd.rf_chain[rf_chain].tx_lut.lut[i];
        }
    }

    x = sx1301ar_conf_board( 0, &cfg_brd );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_conf_board failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }

    printf( "TX Gains: FPGA:%d\n", g.fpga_dig_gain );
    printf( "TX Gains: AD9361: Atten:%d, AuxDAC Vref:%d, AuxDAC:%d\n", g.ad9361_gain.atten, g.ad9361_gain.auxdac_vref, g.ad9361_gain.auxdac_word );
    printf( "TX Gains: AD9361 Temperature Compensation: Tref_code: %u, coeff_a:%d, coeff_b:%d\n", cfg_brd.ad9361_temp_ref, g.ad9361_tcomp.coeff_a, g.ad9361_tcomp.coeff_b );

    /* Configure array */
    x = sx1301ar_start( 1 );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_start failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( 1 );
        return EXIT_FAILURE;
    }

    /* Override SX1301 setup */
    sx1301ar_reg_chip_w( 0, 0, CREG_TX_MODE, 1 ); /* Tx continuous */
    sx1301ar_reg_chip_w( 0, 0, CREG_FSK_TX_GAUSSIAN_SELECT_BT, bt );

    /* Override FPGA setup */
    /* if CW, use signal generator that overrides modulation */
    if( strcmp( mod, "CW" ) == 0 )
    {
        /* Enable signal generator with DC */
        sx1301ar_reg_brd_w( 0, BREG_SIGGEN_FREQ, 0 );
        sx1301ar_reg_brd_w( 0, BREG_TX_SIGGEN_EN, 1 );
    }

    /* Fill Tx packet fields */
    p = sx1301ar_init_tx_pkt( );
    p.freq_hz = f_hz;
    p.tx_mode = TX_IMMEDIATE;
    p.rf_power = rf_power;
    p.rf_chain = rf_chain;
    if( strcmp( mod, "FSK" ) == 0 )
    {
        p.modulation = MOD_FSK;
        p.modrate = br_kbps * 1e3;
    }
    else
    {
        p.modulation = MOD_LORA;
        p.modrate = sx1301ar_sf_nb2enum( sf );
    }
    p.bandwidth = sx1301ar_bw_nb2enum( 1000 * (long)bw_khz );
    p.coderate = CR_4_5;
    p.f_dev = fdev_khz;
    p.preamble = 65535;
    p.invert_pol = false;
    p.no_crc = true;
    p.no_header = true;
    p.size = 1;

    /* Wait for AGC FW to read temperature sensor */
    sx1301ar_wait_ms( 100 );

    /* Send packet */
    x = sx1301ar_send( 0, &p );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_send failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( 0 );
        return EXIT_FAILURE;
    }

    /* Check that modulation is running */
    sx1301ar_wait_us( 5000 );
    x = sx1301ar_tx_status( 0, &tx_status );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_tx_status failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        sx1301ar_stop( 0 );
        return EXIT_FAILURE;
    }
    if( tx_status != TX_EMITTING )
    {
        printf( "ERROR: Modulation not running\n" );
        sx1301ar_stop( 0 );
        return EXIT_FAILURE;
    }

    /* Get SPI control of AD9361 (instead SX1301) */
    sx1301ar_reg_brd_w( 0, BREG_AD9361_SPI_HOST_CTRL, 1 );

    /* Check global attenuation */
    sx1301ar_ad9361_r( 0, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, &u8val );
    sx1301ar_ad9361_r( 0, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, &u8val2 );
    printf( "INFO: final atten used with temperature compensation is %u\n", u8val + (u8val2 << 8) );

    /* waiting for user input */
    printf( "running...\n" );
    printf( "Enter \"n<value>\" to modify atten\n" );
    printf( "Enter \"d<value>\" to modify auxdac_word value\n" );
    printf( "Enter \"a<value>\" to modify temperature compensation coeff A\n" );
    printf( "Enter \"b<value>\" to modify temperature compensation coeff B\n" );
    printf( "Enter \"q\" to quit\n" );
    while( (quit_sig != 1) && (exit_sig != 1) )
    {
        userchar[0] = fgetc( stdin );
        switch( userchar[0] )
        {
        /* Set Atten value */
        case 'n':
            /* Get value from user */
            cret = fgets( userchar, sizeof(userchar), stdin );
            if( cret != NULL )
            {
                g.ad9361_gain.atten = (uint16_t) strtoul( userchar, NULL, 10 );

                /* Set requested attentuation */
                x = ad9361_atten_set_val( 0, rf_chain, g.ad9361_gain.atten, g.ad9361_tcomp.coeff_a, g.ad9361_tcomp.coeff_b, cfg_brd.ad9361_temp_ref );
                if( x < 0 )
                {
                    fprintf( stderr, "ERROR: Failed to set AD9361 atten\n" );
                }
            }
            break;

        /* Set auxdac value */
        case 'd':
            /* Get value from user */
            cret = fgets( userchar, sizeof(userchar), stdin );
            if( cret != NULL )
            {
                g.ad9361_gain.auxdac_word = (uint16_t) strtoul( userchar, NULL, 10 );
                /* Compute Voltage value of DAC */
                if( g.ad9361_gain.auxdac_vref & 0x01 )
                {
                    // Using external Vref = 3.3V
                    val = 2140 + ((g.ad9361_gain.auxdac_word * (3000 - 2140)) / 1024);
                }
                else
                {
                    /* Using external Vref = 2.5V */
                    val = 1670 + ((g.ad9361_gain.auxdac_word * (2500 - 1670)) / 1024);
                }
                printf( "Configuring AD9361 with auxDAC word:%u (estimation: %.2fV)\n",
                        g.ad9361_gain.auxdac_word, ((float) val / 1000) );

                /* Get SPI control of AD9361 (instead SX1301) */
                sx1301ar_reg_brd_w( 0, BREG_AD9361_SPI_HOST_CTRL, 1 );
                if( rf_chain == 1 )
                {
                    u8val = (uint8_t) (g.ad9361_gain.auxdac_word >> 2);
                    sx1301ar_ad9361_w( 0, RREG_AUXDAC_2_WORD, u8val );
                    /* Write LSB part */
                    sx1301ar_ad9361_r( 0, RREG_AUXDAC_2_CONFIG__AUXDAC_2_WORD_LSB, &u8val );
                    u8val = (u8val & 0xFC) + (g.ad9361_gain.auxdac_word & 0x3);
                    sx1301ar_ad9361_w( 0, RREG_AUXDAC_2_CONFIG__AUXDAC_2_WORD_LSB, u8val );
                }
                else
                {
                    u8val = (uint8_t) (g.ad9361_gain.auxdac_word >> 2);
                    sx1301ar_ad9361_w( 0, RREG_AUXDAC_1_WORD, u8val );
                    /* Write LSB part */
                    sx1301ar_ad9361_r( 0, RREG_AUXDAC_1_CONFIG__AUXDAC_1_WORD_LSB, &u8val );
                    u8val = (u8val & 0xFC) + (g.ad9361_gain.auxdac_word & 0x3);
                    sx1301ar_ad9361_w( 0, RREG_AUXDAC_1_CONFIG__AUXDAC_1_WORD_LSB, u8val );
                }
            }
            break;

        /* Set temperature compensation coefficient A value */
        case 'a':
            /* Get value from user */
            cret = fgets( userchar, sizeof(userchar), stdin );
            if( cret != NULL )
            {
                g.ad9361_tcomp.coeff_a = (int16_t) strtoul( userchar, NULL, 10 );

                /* Set requested attentuation */
                x = ad9361_atten_set_val( 0, rf_chain, g.ad9361_gain.atten, g.ad9361_tcomp.coeff_a, g.ad9361_tcomp.coeff_b, cfg_brd.ad9361_temp_ref );
                if( x < 0 )
                {
                    fprintf( stderr, "ERROR: Failed to set temperature compensation coeff A\n" );
                }
            }
            break;

        /* Set temperature compensation coefficient B value */
        case 'b':
            /* Get value from user */
            cret = fgets( userchar, sizeof(userchar), stdin );
            if( cret != NULL )
            {
                g.ad9361_tcomp.coeff_b = (int16_t) strtoul( userchar, NULL, 10 );

                /* Set requested attentuation */
                x = ad9361_atten_set_val( 0, rf_chain, g.ad9361_gain.atten, g.ad9361_tcomp.coeff_a, g.ad9361_tcomp.coeff_b, cfg_brd.ad9361_temp_ref );
                if( x < 0 )
                {
                    fprintf( stderr, "ERROR: Failed to set temperature compensation coeff B\n" );
                }
            }
            break;

        /* Quit application */
        case 'q':
            quit_sig = 1;
            break;

        default:
            break;
        }
    }

    /* Give back SPI control of AD9361 to SX1301 */
    sx1301ar_reg_brd_w( 0, BREG_AD9361_SPI_HOST_CTRL, 0 );

    /* Abort current continuous TX */
    x = sx1301ar_abort_tx( 0 );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_abort_tx failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
    }

    /* Stop the concentrator */
    x = sx1301ar_stop( 1 );
    if( x != 0 )
    {
        printf( "ERROR: sx1301ar_stop failed: %s\n", sx1301ar_err_message( sx1301ar_errno ) );
        return EXIT_FAILURE;
    }
    printf( "INFO: concentrator stopped\n" );

    /* Close SPI link */
    i = spi_linuxdev_close( spi1_context );
    if( i!= 0 )
    {
        printf( "ERROR: closing SPI failed, returned %i\n", i );
        return EXIT_FAILURE;
    }
    printf( "INFO: SPI link closed\n" );

    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/* --- SUBFUNCTIONS DEFINITION ---------------------------------------------- */

static void sig_handler( int sigio )
{
    if( sigio == SIGQUIT )
    {
        quit_sig = 1;
    }
    else if( (sigio == SIGINT) || (sigio == SIGTERM) )
    {
        exit_sig = 1;
    }
}

static int ad9361_atten_set_val( uint8_t brd, uint8_t rf_chain, uint16_t atten, int16_t tcomp_coeff_a, int16_t tcomp_coeff_b, uint8_t ad9361_temp_ref )
{
    uint8_t temperature_code;
    uint16_t atten_compensated;
    uint8_t u8val1, u8val2;
    int32_t atten_check;
    int8_t tcomp_atten;

    /* Check input parameters */
    if( brd >= SX1301AR_MAX_BOARD_NB )
    {
        fprintf( stderr, "ERROR: wrong board number (%u)\n", brd );
        return -1;
    }
    if( (rf_chain != 0) && (rf_chain != 1) )
    {
        fprintf( stderr, "ERROR: wrong rf_chain (%u)\n", rf_chain );
        return -1;
    }

    /* Get current temperature code */
    /* We need to read it directly from the AD9361 radio, and cannot use the
        sx1301ar_get_temperature() function of the HAL, as it won't be updated
        during a running continuous TX */
    if( ad9361_read_temperature( brd, &temperature_code ) != 0 )
    {
        fprintf( stderr, "ERROR: Failed to set attentuation\n" );
        return -1;
    }

    /* Calculate attenuation to be applied to compensate temperature */
    tcomp_atten = sx1301ar_ad9361_temp2atten( temperature_code, ad9361_temp_ref, tcomp_coeff_a, tcomp_coeff_b );

    /* Apply temperature compensation attenuation to given atten (clip[0..255])*/
    atten_check = (int32_t)atten + (int32_t)tcomp_atten;
    if( atten_check < 0 )
    {
        atten_compensated = 0;
    }
    else if( atten_check > AD9361_TXGAIN_ATTEN_MAX )
    {
        /* Max value of AD9361 atten */
        atten_compensated = AD9361_TXGAIN_ATTEN_MAX;
    }
    else
    {
        atten_compensated = (uint16_t)atten_check;
    }

    /* Write attenuation register (9-bits) */
    sx1301ar_ad9361_w( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, (uint8_t)(atten_compensated & 0xFF) );
    sx1301ar_ad9361_w( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, (uint8_t)((atten_compensated >> 8) & 1) );

    /* Read back attenuation register (9-bits) */
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_0 : RREG_TX2_ATTEN_0, &u8val1 );
    sx1301ar_ad9361_r( brd, (rf_chain == 0) ? RREG_TX1_ATTEN_1 : RREG_TX2_ATTEN_1, &u8val2 );

    /* Check attenuation value */
    if( (u8val1 + (u8val2 << 8)) != atten_compensated )
    {
        fprintf( stderr, "ERROR: failed to set attenuation (read:%u)\n", u8val1 + (u8val2 << 8) );
        return -1;
    }
    else
    {
        printf( "INFO: Configuring AD9361 with atten:%d (%.2f dB) (atten_requested:%u atten_tcomp:%d Tref_code:%u Tnow_code:%u, coeff_a:%d, coeff_b:%d)\n", atten_compensated, atten_compensated * 0.25, atten, tcomp_atten, ad9361_temp_ref, temperature_code, tcomp_coeff_a, tcomp_coeff_b );
    }

    return 0;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int ad9361_read_temperature( uint8_t brd, uint8_t * temperature_code )
{
    uint8_t u8val1, u8val2;
    int32_t timeout_ms = 1000; /* 1 second */

    /* Clear bit 0 to prepare for reading (not self-cleared) */
    sx1301ar_ad9361_w( brd, RREG_START_TEMP_READING, 0 );

    /* Start reading temperature */
    sx1301ar_ad9361_w( brd, RREG_START_TEMP_READING, 1 );

    /* Wait for temperature reading to be complete */
    do
    {
        sx1301ar_wait_ms( 100 );
        timeout_ms -= 100;

        sx1301ar_ad9361_r( brd, RREG_START_TEMP_READING, &u8val1 );
        u8val2 = (TAKE_N_BITS_FROM( u8val1, 1, 1 )); /* bit 1 is set to 1 when reading complete */
    } while( (u8val2 != 1) && (timeout_ms > 0) );

    if( (timeout_ms <= 0) && (u8val2 != 1) )
    {
        printf( "ERROR: Failed to read current temperature, timeout\n" );
        return -1;
    }

    /* Get current temperature */
    sx1301ar_ad9361_r( brd, RREG_TEMPERATURE, temperature_code );

    return 0;
}

int extract_rfchain_configuration( JSON_Array * conf_rfchain_array, sx1301ar_board_cfg_t *cfg_brd )
{
    int i, j;
    JSON_Object * conf_txgain_obj;
    JSON_Object * conf_rfchain_obj;
    JSON_Array * conf_txlut_array;
    JSON_Value * val;

    if( conf_rfchain_array != NULL )
    {
        for( i = 0; i < (int) json_array_get_count( conf_rfchain_array ); i++ )
        {
            /* Sanity check */
            if( i >= SX1301AR_BOARD_RFCHAIN_NB )
            {
                printf( "ERROR: rf_chain %d not supported, skip it\n", i );
                return EXIT_FAILURE;
            }

            /* Get RF chain conf object */
            conf_rfchain_obj = json_array_get_object( conf_rfchain_array, i );

            /* Parse configuration */
            cfg_brd->rf_chain[i].rx_enable = (bool) json_object_get_boolean( conf_rfchain_obj, "rx_enable" );
            cfg_brd->rf_chain[i].tx_enable = (bool) json_object_get_boolean( conf_rfchain_obj, "tx_enable" );

            /* set configuration for tx gains */
            conf_txlut_array = json_object_get_array( conf_rfchain_obj, "tx_lut" );
            if( conf_txlut_array != NULL )
            {
                cfg_brd->rf_chain[i].tx_lut.size = json_array_get_count( conf_txlut_array );
                printf( "INFO: rf_chain %d: Configuring Tx LUT with %u indexes\n", i, cfg_brd->rf_chain[i].tx_lut.size );
                for( j = 0; j < (int) cfg_brd->rf_chain[i].tx_lut.size; j++ )
                {
                    /* Sanity check */
                    if( j >= SX1301AR_BOARD_MAX_LUT_NB )
                    {
                        printf( "ERROR: TX LUT index %d not supported, skip it\n", j );
                        break;
                    }

                    /* Get TX gain object from LUT */
                    conf_txgain_obj = json_array_get_object( conf_txlut_array, j );

                    /* rf power */
                    val = json_object_dotget_value( conf_txgain_obj, "rf_power" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].rf_power = (int8_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "rf_power", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].rf_power = 0;
                    }

                    /* FPGA dig gain */
                    val = json_object_dotget_value( conf_txgain_obj, "fpga_dig_gain" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].fpga_dig_gain = (uint8_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "fpga_dig_gain", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].fpga_dig_gain = 0;
                    }

                    /* AD9361 atten */
                    val = json_object_get_value( conf_txgain_obj, "ad9361_atten" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.atten = (uint16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_atten", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.atten = 0;
                    }

                    /* AD9361 auxdac vref */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_auxdac_vref" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_vref = (uint8_t) json_value_get_number( val );
                    }
                    else
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_vref = 0;
                    }

                    /* AD9361 auxdac word */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_auxdac_word" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_word = (uint16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_auxdac_word", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_gain.auxdac_word = 0;
                    }

                    /* AD9361 temperature compensation coeff A */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_tcomp_coeff_a" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_a = (int16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_tcomp_coeff_a", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_a = 0;
                    }

                    /* AD9361 temperature compensation coeff B */
                    val = json_object_dotget_value( conf_txgain_obj, "ad9361_tcomp_coeff_b" );
                    if( json_value_get_type( val ) == JSONNumber )
                    {
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_b = (int16_t) json_value_get_number( val );
                    }
                    else
                    {
                        printf( "WARNING: Data type for %s[%d] seems wrong, please check\n", "ad9361_tcomp_coeff_b", j );
                        cfg_brd->rf_chain[i].tx_lut.lut[j].ad9361_tcomp.coeff_b = 0;
                    }
                }
            }
        }
    }
    else
    {
        printf( "WARNING: no rf_chain configured\n" );
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

static int spi1_read( uint8_t header, uint16_t address, uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_read( header, spi1_context, address, data, size, status );
}

static int spi1_write( uint8_t header, uint16_t address, const uint8_t * data, uint32_t size, uint8_t * status )
{
    return spi_linuxdev_write( header, spi1_context, address, data, size, status );
}

/* --- EOF ------------------------------------------------------------------ */
